/*
 * Copyright (C) 2011 Nokia Corporation. All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 * Markku Savela
 */

/**
 * @file SecureJournal.cpp
 *
 * Maintain secure journal file
 *
 * This program attempts to provide a light journaling class
 * using a secret from BB5 and HMAC authentication algorithm
 */

#include <stddef.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <errno.h>
#include <err.h>
#include <openssl/sha.h>

#include "SecureJournal.h"

// Open the journal file, only exists to collect
// the opening mode into one place...
static int open_journal(const char *name)
	{
	return open(name, O_RDWR|O_CREAT, 0600);
	}

static unsigned char *open_and_map_file(const char *name, size_t &size, int &fd)
	{
	unsigned char *mapped;

	size = 0;
	fd = open_journal(name);
	if (fd == -1)
		return NULL;

	struct stat info;
	if (fstat(fd, &info))
		goto badfile;

	size = info.st_size;
	mapped = (unsigned char *)mmap(NULL, size, PROT_READ, MAP_PRIVATE, fd, 0);
	if (mapped != MAP_FAILED)
		return mapped;

badfile:
	size = 0;
	return NULL;
	}

static void unmap_file(unsigned char *mapped, size_t maplen)
	{
	if (mapped)
		munmap(mapped, maplen);
	}



#if HAVE_AEGIS_CRYPTO_H

// Implement Journal using Aegis Crypto
//
// The format of the journal file:
//
// size_t
//	the length of the key material = N
// char[N]
//	the key material (N bytes)
// char[M] (M >= 0)
//	variable length journal content
// char[TRUNCATED_LENGTH]
//	the HMAC digest value (base on ALL preceding content)
//
// NOTE: This is a binary format, only understandable by this code
// that writes it. This is not portable (nor intended to be) to other
// programs or hosts.

#include <aegis_crypto.h>
#include <openssl/sha.h>

#define TRUNCATED_LENGTH	(SHA256_DIGEST_LENGTH-8)

class Journal : public SecureJournal
	{
public:
	Journal(const char *name, const char *cred)
		: name(name), cred(cred), fd(-1), mapped(NULL), state(1) {}
	virtual ~Journal();
	virtual const unsigned char *content(size_t &size);
	virtual void release();
	virtual int append(const void *data, size_t len);
private:
	const char *const name;
	const char *const cred;

	int fd;
	unsigned char *mapped;
	size_t maplen;

	// Journal state:
	// 1, journal not started yet
	// 0, journal write possible
	// -1, error no writing occurs
	int state;
	SHA256_CTX ctx;
	// The opad is sensitive information and if paranoid, should
	// not be kept around. Instead, should be recovered via
	// aegis_crypto_decrypt each time used, causing extra
	// overhead... When kept around like this, the security
	// depends on that unauthorized eyes cannot read this from
	// more (=> careful with core dumps, no memory mapping and no
	// ptracing?)
	unsigned char opad[SHA256_CBLOCK];
	};

const unsigned char *Journal::content(size_t &size)
	{
	SHA256_CTX tmp;
	unsigned char ipad[SHA256_CBLOCK];
	unsigned char digest[SHA256_DIGEST_LENGTH];
	RAWDATA_PTR key = NULL; // Holds DECTRYPTED key (plain)
	size_t bytes, len;
	aegis_crypto_result res;
	unsigned char *p;
	
	p = mapped = open_and_map_file(name, size, fd);
	if (fd < 0)
		{
		// Must be able to open the file at least -- no
		// journal will be generated if file open failed.
		state = -1;
		goto no_journal;
		}
	// Extract and verify the fixed crypto header and trailer size
	if (size < sizeof(size_t) + TRUNCATED_LENGTH)
		{
		if (size)
			warnx("Journal file length %u bytes is short", size);
		goto no_journal; // Not large enough even for fixed fields!
		}
	size -= sizeof(size_t) + TRUNCATED_LENGTH;
	bytes = *(size_t *)p;
	p += sizeof(size_t);
	if (bytes > size)
		{
		warnx("Journal file %u bytes is short for required crypto info %u", size, bytes);
		goto no_journal; // Not large enough for crypto material
		}
	size -= bytes; // == plain message length now

	// Extract variable crypto information
	res = aegis_crypto_decrypt(p, bytes, cred, &key, &len);
	if (res)
		{
		warnx("Journal key decryption failed (%d)", res);
		goto no_journal; // Cannot decrypt key
		}
	if (len != sizeof(ipad))
		{
		warnx("Journal key decryption returned wrong length %u != %u", bytes, sizeof(ipad));
		goto no_journal; // Cannot decrypt key
		}
	memcpy(ipad, key, sizeof(ipad));
	memcpy(opad, key, sizeof(opad));
	memset(key, 0, len); // scrub!
	aegis_crypto_free(key);
	key = NULL;

	for (size_t i = SHA256_DIGEST_LENGTH; i < sizeof(ipad); ++i)
		if (ipad[i])
			{
			warnx("Journal key decryption failed -- not zeropadded");
			goto no_journal;
			}
	for (size_t i = 0; i < sizeof(ipad); ++i)
		{
		ipad[i] ^= 0x36;
		opad[i] ^= 0x5c;
		}

	SHA256_Init(&ctx);
	SHA256_Update(&ctx, ipad, sizeof(ipad));
	memset(ipad, 0, sizeof(ipad)); // not needed after this, scrub
	// HMAC computed over the whole file...
	SHA256_Update(&ctx, mapped, sizeof(size_t) + bytes + size);

	// ..if Final changes ctx, need to save it!
	tmp = ctx;
	SHA256_Final(digest, &tmp);

	SHA256_Init(&tmp);
	SHA256_Update(&tmp, opad, sizeof(opad));
	SHA256_Update(&tmp, digest, sizeof(digest));
	SHA256_Final(digest, &tmp);

	p += bytes;
	if (memcmp(digest, p + size, TRUNCATED_LENGTH) == 0)
		{
		state = 0;
		return p;
		}

	// Authentication failure -- ignore old journal
		{
		// Try to save the bad version for study, if needed.
		const char *bad = ".bad";
		char bad_name[strlen(name)+strlen(bad)+1];
		strcpy(bad_name, name);
		strcat(bad_name, bad);
		rename(name, bad_name);
		close(fd);
		fd = -1;
		}
	warnx("Journal HMAC authentication failed");
no_journal:
	if (key)
		// no scrubbing needed, key is not used!
		aegis_crypto_free(key);
	size = 0;
	return NULL;
	}

int Journal::append(const void *data, size_t len)
	{
	SHA256_CTX tmp;
	unsigned char digest[SHA256_DIGEST_LENGTH];

	if (state < 0)
		return 0; // Error state, skip everything
	if (state)
		{
		// The creation of new journal delayed until it really is
		// needed...

		size_t bytes;
		aegis_crypto_result res;
		RAWDATA_PTR key = NULL; // holds ENCRYPTED key!
		unsigned char ipad[SHA256_CBLOCK];

		// opad will contain sensitive data, try to keep it
		// away from swap, but don't get upset, if mlock fails
		// (cap_ipc_lock needed) -- ignore error.
		//
		// NOTE: This lock is not released on destructor --
		// beware, if you use multiple instances of this on a
		// long running process. (Reason: mlock's don't stack
		// and without more complex code, we don't know how
		// many locks we have on affected page!).
		mlock(&opad, sizeof(opad));
		
		// assumed SHA256_CBLOCK <= SHA256_DIGEST_LENGTH !!!
		memset(opad, 0, sizeof(opad));
		bytes = aegis_crypto_random(opad, SHA256_DIGEST_LENGTH);
		if (bytes != SHA256_DIGEST_LENGTH)
			{
			warnx("Journal random bit generation returend %u bytes != %u",
				bytes, SHA256_DIGEST_LENGTH);
			goto no_journal;
			}
		res = aegis_crypto_encrypt(opad, sizeof(opad), cred, &key, &bytes);
		if (res)
			{
			warnx("Journal key encryption failed (%d) returned bytes %u", res, bytes);
			goto no_journal;
			}

		// Setup HMAC pads using the secret
		memcpy(ipad, opad, sizeof(ipad));
		for (size_t i = 0; i < sizeof(ipad); ++i)
			{
			ipad[i] ^= 0x36;
			opad[i] ^= 0x5c;
			}
		
		SHA256_Init(&ctx);
		SHA256_Update(&ctx, ipad, sizeof(ipad));
		memset(ipad, 0, sizeof(ipad)); // not needed after this, scrub
		SHA256_Update(&ctx, &bytes, sizeof(bytes));
		SHA256_Update(&ctx, key, bytes);
		
		// Open the file, if not yet opened (happens only if
		// content() was not called).
		if (fd < 0)
			fd = open_journal(name);

		// Make empty journal, ready to be appended
		state =	-(lseek(fd, 0, SEEK_SET) == (off_t)-1 ||
			ftruncate(fd, 0) != 0 ||
			(size_t)write(fd, &bytes, sizeof(bytes)) != sizeof(bytes) ||
			(size_t)write(fd, key, bytes) != bytes);
		aegis_crypto_free(key);
		}
	else
		state = -(lseek(fd, -TRUNCATED_LENGTH, SEEK_END) == (off_t)-1);

	SHA256_Update(&ctx, data, len);
	
	// If final changes ctx, need to save it!
	tmp = ctx;
	SHA256_Final(digest, &tmp);

	SHA256_Init(&tmp);
	SHA256_Update(&tmp, opad, sizeof(opad));
	SHA256_Update(&tmp, digest, sizeof(digest));
	SHA256_Final(digest, &tmp);

	if (!state)
		{
		state =	-((size_t)write(fd, data, len) != len ||
			write(fd, digest, TRUNCATED_LENGTH) != TRUNCATED_LENGTH ||
			fdatasync(fd) != 0);
		}
	return state;
no_journal:
	state = -1;
	return state;
	}

Journal::~Journal()
	{
	memset(opad, 0, sizeof(opad)); // scrub!
	unmap_file(mapped, maplen);
	if (fd >= 0)
		close(fd);
	}

#else
// Implement Journal using plain file (unsecure -- for debugging only)

class Journal : public SecureJournal
	{
public:
	Journal(const char *name, const char *cred)
		: name(name), cred(cred), fd(-1), mapped(NULL) {}
	virtual ~Journal();
	virtual const unsigned char *content(size_t &size);
	virtual void release();
	virtual int append(const void *data, size_t len);
private:
	const char *const name;
	const char *const cred;
	
	int fd;
	unsigned char *mapped;
	size_t maplen;
	};

Journal::~Journal()
	{
	unmap_file(mapped, maplen);
	if (fd >= 0)
		close(fd);
	}

const unsigned char *Journal::content(size_t &size)
	{
	mapped = open_and_map_file(name, size, fd);
	maplen = size;
	lseek(fd, 0, SEEK_END);
	return mapped;
	}

int Journal::append(const void *data, size_t len)
	{
	size_t actual = write(fd, data, len);
	fdatasync(fd);
	return -(actual != len); // -1 on error, 0 otherwise
	}

#endif

void Journal::release()
	{
	unmap_file(mapped, maplen);
	mapped = NULL;
	}

SecureJournal *SecureJournal::New(const char *path, const char *cred)
	{
	return path && cred ? new Journal(path, cred) : NULL;
	}

