/*
 * Copyright (C) 2011 Nokia Corporation. All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 * Markku Savela
 */

#ifndef _SECURE_JOURNAL_H_
#define _SECURE_JOURNAL_H_

/**
 * @file SecureJournal.h
 *
 * Define the header for secure journal class
 */
#include <sys/types.h>

class SecureJournal
	{
public:
	/**
	 * Return current content of the journal
	 *
	 * @size The length of the content, 0 if none
	 * @return Start of the content or NULL (size==0)
	 *
	 * Open the journal file and if there is previously
	 * stored information, returns the content.
	 *
	 * Returns no content, if it does not authenticate with the
	 * provided credential token (even if the content would
	 * authenticate with some other token). The authentication
	 * requires that the callers task context includes the token.
	 *
	 * Calling this also sets up the journal file so that
	 * subsequent append calls append to the old content. If
	 * content is not called, the first append call makes the
	 * journal empty, before appending.
	 *
	 * The returned address is valid until instance
	 * deleted or release() called.
	 */
	virtual const unsigned char *content(size_t &size) = 0;
	/**
	 * Release the old content address
	 *
	 * Release resource bound to the old content returned by the
	 * content() function. After calling this, the address
	 * returned by content() is not valid.
	 */
	virtual void release() = 0;
	/**
	 * Append data to the journal
	 *
	 * @data Start of the data
	 * @len Length of the data
	 *
	 * Append new data to the journal file.
	 *
	 * If content() has not been called, the first append
	 * truncates the journal to empty, deleting the old journal
	 * content.
	 */
	virtual int append(const void *data, size_t len) = 0;

	/**
	 * Create a new instance of SecureJournal
	 * @path The filepath of the journal file
	 * @cred The credential token
	 * 
	 * Create an instance of the SecureJournal class. This only
	 * initializes the structure and does not yet touch or look at
	 * the journal file.
	 *
	 * The specified credential token must be included in the
	 * callers task context, or journalling does not happen.
	 *
	 * This does not copy the path nor cred, and both strings must
	 * remain valid until destruction of the class.
	 */
	static SecureJournal *New(const char *path, const char *cred);
	/**
	 * Destructor
	 */
	virtual ~SecureJournal() {}
	};
#endif
