/*
 * This file is part of AEGIS
 *
 * Copyright (C) 2009-2010 Nokia Corporation
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2.1 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 *
 * Author: Markku Savela
 */
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <err.h>
#include <ctype.h>
#include <sys/creds.h>
#include <pwd.h>
#include <grp.h>
#include <aegis-session.h>

// Local prototypes
static int load_aegis_session(void);
static int resolve_login_creds(creds_t * creds, struct passwd *login);

/*
 * Compute a new set of credentials and run the
 * executable under modified credentials.
 *
 * Without special privileges, this can only
 * decrease the current credentials.
 *
 * aegis-exec [options] [path [args]]
 *
 * Options are:
 *  -c        Clear credentials
 *  -p pid    Load credentials from a task
 *  -n        Don't load credentials
 *  -a cred   Add the credential
 *  -d cred   Delete the credential
 *  -m flags  Modifier flags (use quotes)
 *  -h        This text
 *  -D        Print current state of credentials set
 *  -u user   Identity change to specified user.
 *  -l        Simulate login shell.
 *  -s        Load Aegis Session
 */
int main(int argc, char **argv)
{
	int do_creds = 1;
	int do_aegis_session = 0;
	int do_login_shell = 0;
	creds_t creds = creds_init();
	struct passwd *login = NULL;
	const char *arg;
	int flags, mask;
	const char *modifiers = NULL;
	int ret;

	flags = credp_str2flags(modifiers, &mask);

	while ((arg = *++argv) != NULL && --argc > 0) {
		int option;
		const char *str;
		creds_value_t value;
		creds_type_t type;
		int index;
		char buf[200];

		if (*arg != '-')
			break;
		++arg;
		while ((option = *arg++) != 0) {
			int sub = 0;
			switch (option) {
			case 'n':
				do_creds = 0;
				break;
			case 'c':
				/* Clear current credentials set */
				creds_clear(creds);
				break;
			case 'p':
				/* Reload creds from a task */
				str = arg;
				if (*str == 0 && argc > 1) {
					str = *++argv;
					--argc;
				}
				option = atoi(str);
				creds_free(creds);
				creds = creds_gettask(option);
				goto next_arg;
			case 'm':
				modifiers = arg;
				if (*modifiers == 0 && argc > 1) {
					modifiers = *++argv;
					--argc;
				}
				flags = credp_str2flags(modifiers, &mask);
				if (flags == -1) {
					if (mask) {
						warnx("Conflicting modifier in '%s'\n", modifiers);
					} else {
						warnx("Unknown modifier in '%s'\n", modifiers);
					}
					exit(EXIT_FAILURE);
				}
				goto next_arg;
			case 'd':
				++sub;
				/* FALL THROUGH */
			case 'a':
				/* Add credential to the set */
				str = arg;
				if (*str == 0 && argc > 1) {
					/* The next arg is the creds string */
					str = *++argv;
					--argc;
				}
				type = creds_str2creds(str, &value);

				if (type == CREDS_BAD) {
					warnx("Cannot translate '%s' into credential value\n", str);
					exit(EXIT_FAILURE);
				}

				if (sub)
					creds_sub(creds, type, value);
				else if (creds_add(&creds, type, value)) {
					warnx("Couldn't add credential '%s' (%d,%ld) into credentials\n",
					      str, type, value);
					exit(EXIT_FAILURE);
				}
				goto next_arg;
			case 'D':
				credp_flags2str(flags, ~0, buf, sizeof(buf));
				printf("\nCurrent policy: %s\n", buf);
				for (index = 0; (type = creds_list(creds, index, &value)) != CREDS_BAD; ++index) {
					(void)creds_creds2str(type, value, buf, sizeof(buf));
					buf[sizeof(buf)-1] = 0;
					printf("%s\n", buf);
				}
				break;
			case 'u':
				str = arg;
				if (*str == 0 && argc > 1) {
					/* The next arg is the creds string */
					str = *++argv;
					--argc;
				}

				login = getpwnam(str);
				if (!login) {
					warnx("Cannot find user '%s' from system\n", str);
					exit(EXIT_FAILURE);
				}

				ret = resolve_login_creds(&creds, login);

				if (ret < 0) {
					warnx("Failed to prepare credentials for user '%s'\n", str);
					exit(EXIT_FAILURE);
				}
				goto next_arg;
			case 'l':
				do_login_shell = 1;
				break;
			case 's':
				do_aegis_session = 1;
				break;
			default:
				warnx("Unkonwn option '%c'\n", option);
				/* FALL THROUGH */
			case 'h':
				warnx("Usage: aegis-exec [options] [path] [args]]\n"
					" Options are:\n"
					"  -c        Clear credentials\n"
					"  -p pid    Load credentials from a task\n"
					"  -n        Don't load credentials\n"
					"  -a cred   Add the credential\n"
					"  -d cred   Delete the credential\n"
					"  -m flags  Modifier flags (use quotes)\n"
					"  -h        This text\n"
					"  -D        Print current state of credentials set\n"
					"  -u user   Identity change to specified user.\n"
					"  -l        Simulate login shell.\n"
					"  -s        Load Aegis Session\n");
				exit(EXIT_FAILURE);
			}
		}
	next_arg:
		;
	}

	// Only allow login shell if user was specified
	if (do_login_shell && login == NULL) {
		warnx("Couldn't do login shell without specifying user\n");
		exit(EXIT_FAILURE);
	}

	// Setup credentials for process
	if (do_creds && creds_set2(creds, flags)) {
		warnx("Couldn't set the credentials for the current process\n");
		exit(EXIT_FAILURE);
	}

	// Simulate login shell; prepare environment variables
	if (login != NULL) {
		int res;

		// Change to user's home directory
		res = chdir(login->pw_dir);
		if (res < 0) {
			res = chdir("/");
		}

		if (do_login_shell) {
			const char *term;

			// Leave TERM in place
			term = getenv("TERM");

			// Start with scratch
			clearenv();

			setenv("TERM", term, 1);

			// Use schematics of busybox for PATH
			if (login->pw_uid != 0) {
				// Simulate non-root PATH
				setenv("PATH", "/bin:/usr/bin", 1);
			} else {
				// Simulate root PATH
				setenv("PATH", "/sbin:/usr/sbin:/bin:/usr/bin", 1);
			}
		}

		// Prepare standard user specific variables
		setenv("USER", login->pw_name, 1);
		setenv("LOGNAME", login->pw_name, 1);
		setenv("HOME", login->pw_dir, 1);
		setenv("SHELL", login->pw_shell, 1);
	}

	// Load session data from Aegis Session
	if (do_aegis_session) {
		if (!load_aegis_session()) {
			exit(EXIT_FAILURE);
		}
	}

	// Finally execute application
	if (arg || do_login_shell) {
		if (do_login_shell) {
			// For login shell version use shell to launch application
			char ** new_argv;
			char *app_name;
			int new_argc;
			int app_argc;
			int i, j;

			// Prepare arguments for shell launching

			// Count args...
			for (app_argc = 0; argv[app_argc] && *argv[app_argc]; app_argc++) {

			}

			// Determine amount of new arguments for shell launching
			new_argc = 1 /* argv0 */ + 1 /* -c */ + app_argc /* args */ + 1 /* NULL */;

			// Allocate memory for new argv
			new_argv = (char **)malloc(sizeof(char *) * new_argc);
			if (!new_argv) {
				warnx("Failed to allocate memory\n");
				exit(EXIT_FAILURE);
			}

			// Prepare shell name for argv[0]
			app_name = strrchr(login->pw_shell, '/');
			if (app_name) {
				int len;
				char *tmp;

				app_name++;

				len = snprintf(NULL, 0, "-%s", app_name);

				tmp = malloc(len + 1);
				if (!tmp) {
					warnx("Failed to allocate memory\n");
					exit(EXIT_FAILURE);
				}

				snprintf(tmp, len + 1, "-%s", app_name);

				app_name = tmp;
			} else {
				app_name = login->pw_shell;
			}

			i = 0;

			new_argv[i++] = app_name;

			// Only prepare shell command execution if there is command to execute
			if (arg) {
				// -c is used to execute command and exit after that
				new_argv[i++] = "-c";

				// Use existing args for shell arguments
				for (j = 0; j < app_argc; j++) {
					new_argv[i++] = argv[j];
				}
			}

			// Make sure list is terminated
			new_argv[i++] = NULL;

			// Finally launch the shell
			execvp(login->pw_shell, new_argv);
		} else {
			// Without login shell approach; just launch the application directly
			execvp(arg, argv);
		}

		// exec()'s only return on failure
		perror(arg);
		exit(EXIT_FAILURE);
	}
	exit(EXIT_SUCCESS);
}

/**
 * Loads Aegis Session for executed application.
 *
 * @return Non-zero on success, zero on failure.
 */
static int load_aegis_session(void)
{
	AegisSessionEntry *sessionData = NULL;
	AegisSessionEntry *tmp;
	int ret;

	ret = aegisSessionOpen(&sessionData, 0);
	if (!ret) {
		return 0;
	}

	tmp = sessionData;

	while(tmp) {
		if (tmp->type == EnvironmentVariable) {
			ret = setenv(tmp->name, tmp->value, 1);
			if (ret < 0) {
				perror("setenv()");

				aegisSessionFreeData(sessionData);
				sessionData = NULL;

				(void) aegisSessionClose();

				return 0;
			}
		} else if (tmp->type == UlimitResourceLimit) {
			rlim_t soft, hard;

			if (aegisSessionTranslateUlimit(tmp, &soft, &hard)) {
				if (strcmp(tmp->name, "ulimit.stack_size") == 0) {
					// Configure stack size for process
					struct rlimit cur_limits;

					ret = getrlimit(RLIMIT_STACK, &cur_limits);
					if (ret < 0) {
						perror("getrlimit");
						
						aegisSessionFreeData(sessionData);
						sessionData = NULL;

						(void) aegisSessionClose();

						return 0;
					}

					// Convert units from KiB to bytes.					
					if (soft != RLIM_INFINITY) {
						soft *= 1024;
					}

					if (hard != RLIM_INFINITY) {
						hard *= 1024;
					}

					// Value sanity check
					if (soft > hard) {
						soft = hard;
					}

					// Setup new values
					cur_limits.rlim_cur = soft;
					cur_limits.rlim_max = hard;

					ret = setrlimit(RLIMIT_STACK, &cur_limits);
					if (ret < 0) {
						perror("setrlimit");
						
						aegisSessionFreeData(sessionData);
						sessionData = NULL;

						(void) aegisSessionClose();

						return 0;
					}
				} else if (strcmp(tmp->name, "ulimit.max_memory_size") == 0) {
					// Configure max memory size for process
					struct rlimit cur_limits;

					ret = getrlimit(RLIMIT_RSS, &cur_limits);
					if (ret < 0) {
						perror("getrlimit");
						
						aegisSessionFreeData(sessionData);
						sessionData = NULL;

						(void) aegisSessionClose();

						return 0;
					}

					// Convert units from KiB to bytes.					
					if (soft != RLIM_INFINITY) {
						soft *= 1024;
					}

					if (hard != RLIM_INFINITY) {
						hard *= 1024;
					}

					// Value sanity check
					if (soft > hard) {
						soft = hard;
					}

					// Setup new values
					cur_limits.rlim_cur = soft;
					cur_limits.rlim_max = hard;

					ret = setrlimit(RLIMIT_RSS, &cur_limits);
					if (ret < 0) {
						perror("setrlimit");
						
						aegisSessionFreeData(sessionData);
						sessionData = NULL;

						(void) aegisSessionClose();

						return 0;
					}
				}
			}
		}

		tmp = tmp->next;
	}

	aegisSessionFreeData(sessionData);
	sessionData = NULL;

	(void) aegisSessionClose();

	return 1;
}

static int resolve_login_creds(creds_t * creds, struct passwd *login)
{
	struct group *grp;
	char **members;
	int ret;

	long uid_type;
	long gid_type;
	long grp_type;

	// Resolve credential types
	uid_type = creds_str2creds("UID::", NULL);
	if (uid_type == CREDS_BAD) {
		return -1;
	}

	gid_type = creds_str2creds("GID::", NULL);
	if (gid_type == CREDS_BAD) {
		return -1;
	}

	grp_type = creds_str2creds("GRP::", NULL);
	if (grp_type == CREDS_BAD) {
		return -1;
	}

	// Configure UID
	ret = creds_add(creds, uid_type, login->pw_uid);
	if (ret < 0) {
		return -1;
	}

	// Configure GID
	ret = creds_add(creds, gid_type, login->pw_gid);
	if (ret < 0) {
		return -1;
	}

	// Add GID also as supplementary group
	ret = creds_add(creds, grp_type, login->pw_gid);
	if (ret < 0) {
		return -1;
	}

	// Figure out supplementary groups for user
	setgrent();

	grp = getgrent();

	while (grp != NULL) {
		// Check if group has user listed
		members = grp->gr_mem;

		while(*members) {
			if (strcmp(*members, login->pw_name) == 0) {
				// User found; add group as supplementary group
				ret = creds_add(creds, grp_type, grp->gr_gid);
				if (ret < 0) {
					return -1;
				}

				break;
			}

			members++;
		}

		// Get next group
		grp = getgrent();
	}

	endgrent();

	return 0;
}
