/*
 * This file is part of AEGIS
 *
 * Copyright (C) 2009 Nokia Corporation
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2.1 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 *
 * Author: Markku Savela
 */

#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <err.h>
#include <errno.h>
#include "librestok.h"

#define MERGE_BLOB_SIZE 4000

// This is a simple test program which reads the resource
// configuration and attempts to load it to the kernel
// module.
// This should fail, if not run as root.
// This should succeed, if run as root.
//
// (in final secured system, running this are root
// should not work either).
//
int main(int argc, char **argv)
	{
	const char *config = NULL;
	const char *merge = NULL;
	size_t merge_len = 0;
	int print_only = 0;
	int untrusted = 0;

	const char *arg;

	// Process the command line
	while ((arg = *++argv) != NULL && --argc > 0)
		{
		int option;

		if (*arg != '-')
			break;
		++arg;
		while ((option = *arg++) != 0)
			{
			switch (option)
				{
				case 'c':
					config = arg;
					if (*config == 0 && argc > 1)
						{
						config = *++argv;
						--argc;
						}
					goto next_arg;

				case 'u':
					untrusted = 1;
					break;

				case 'm':
					merge = arg;
					if (*merge == 0 && argc > 1)
						{
						merge = *++argv;
						--argc;
						}
					goto next_arg;

				case 'p':
					print_only = 1;
					break;

				case 'r':
					print_only = 2;
					break;

				default:
					warnx("Unkonwn option '%c'\n", option);
					warnx("Usage: aegis-loader [options]\n"
						" Possible options are:\n"
						"  -c file     Alternate configuration file\n"
						"  -u          Untrusted configuration\n"
						"  -m [-|file] Merge update from stdin or file\n"
						"  -p          Only print the parsed configuration\n");
					exit(EXIT_FAILURE);
				}
			}
		next_arg:
		;
		}

	// With '-m' or '-m -', read the "merge section" from stdin
	if (merge && (*merge == 0 || strcmp(merge,"-") == 0))
		{
		size_t size = 0;
		size_t used = 0;
		char *buf = NULL;
		int ret;

		while(1)
			{
			if (used == size)
				{
				// Buffer is fully used -> allocate more memory for it
				size += MERGE_BLOB_SIZE;
				buf = (char *)realloc(buf, size);
				if (buf == NULL)
					{
					warn("Merge buffer allocation (%d) failed", size);
					// ..just ignore memory leaks--exit cleans
					exit(EXIT_FAILURE);
					}
				}

			ret = read(STDIN_FILENO, buf + used, size - used);
			if (ret < 0)
				{
				if (ret == -EINTR)
					{
					// read() was interrupted by signal, just try again.
					}
				else
					{
					// All other errors codes leads to the failure.
					warn("While reading merge input");
					exit(EXIT_FAILURE);
					}
				}
			else if (ret == 0)
				{
				// Input was completely processed.
				break;
				}
			else
				{
				used += ret;
				}

			}

		merge = buf;
		merge_len = used;
		}
	
	// Load the configuration into memory.
	const void *cnf = restok_cnf_load(config, merge, merge_len);

	if (merge_len && merge)
		free((void *)merge);

	if (cnf == NULL)
		{
		warnx("Could not load the configuration");
		exit(EXIT_FAILURE);
		}

	if (untrusted)
		restok_cnf_mark_untrusted(cnf);
	else
		untrusted = restok_cnf_is_untrusted(cnf);

	if (print_only)
		{
		(print_only > 1 ? restok_cnf_dump_raw : restok_cnf_dump)(cnf);
		exit(EXIT_SUCCESS);
		}

	// Become a "resource manager", so that configuration can
	// loaded into the kernel (only succeeds with root).
	if (restok_be_manager())
		{
		//warnx("Could not become a manager -- will try to load the conf anyway");
		// This means, kernel is not supporting Aegis, no
		// point continuing...
		exit(EXIT_SUCCESS);
		}

	// Try to define the dynamic resource tokens for the kernel module.
	if (restok_cnf_define(cnf))
		warnx("Apparently failed to define the resources for the kernel module");

	if (restok_cnf_activate(cnf))
		{
		warnx("Failed activating the policies");
		}

	// Release the configuration data.
	restok_cnf_free(cnf);
	}
