/*
 * This file is part of AEGIS
 *
 * Copyright (C) 2009-2010 Nokia Corporation
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2.1 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 *
 * Author: Markku Savela
 */

#include <stddef.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <grp.h>
#include <errno.h>
#include <err.h>
#include <sys/creds.h>

#ifdef HAVE_LINUX_AEGIS_CREDP_H
extern "C"
	{
#	include <linux/aegis/credp.h>
	}
#endif

#include "librestok.h"
#include "restok_def.h"
#include "restok_cnf.h"
#include "socketpolicy.h"

/* A temporary fallback kludge, which can be removed later:
   The default root of kernel module 'restok' has been changed
   from '_system_' to '' (empty string). Change can be detected
   by presence of RESTOK_DEFAULT_ROOT symbol in kernel header.
*/
#ifndef RESTOK_DEFAULT_ROOT
#define RESTOK_DEFAULT_ROOT ""
#endif

using namespace RESTOK;

static const creds_type_t type_GRP = creds_str2creds("GRP::", NULL);

static const struct configuration* load_cnf(const char *path, const char *update, size_t update_len)
	{
	// Check that type_GRP has been properly resolved (give only one warning)
	if (type_GRP == CREDS_BAD)
		warnx("Couldn't find credentials type for 'GRP::'");

	return configuration::load(path, update, update_len);
	}

static creds_type_t resolve_resource(const char *prefix, const char *name, creds_value_t& value)
	{
	// ..this copying is ugly... would need str2creds service where
	// prefix and name are separated!
	char tmp[512];
	const int res = snprintf(tmp, sizeof(tmp), "%s::%s", prefix, name);
	if (res < 0 || res >= (int)sizeof(tmp))
		{
		value = CREDS_BAD;
		return CREDS_BAD;
		}
	return (creds_type_t)creds_str2creds(tmp, &value);
	}

static int load_object_creds(const char *prefix, int flags, const char *path, creds_t creds)
	{
	int res = strlen(path);
	char tmp[512];

	if (prefix)
		{
		// ..this copying is ugly... would need load service where
		// prefix or namespace and name are separated!
		res = snprintf(tmp, sizeof(tmp), "%s::%s", prefix, path);
		if (res < 0 || res >= (int)sizeof(tmp))
			return -1;
		path = tmp;
		}
	//warnx("Loading creds for path '%s'", path);
#ifdef HAVE_LINUX_AEGIS_CREDP_H
	res = creds_load(flags, path, creds);
#else
	res = -EPERM;
#endif
	return res;
	}

static long register_resource(const char *name, long owner)
	{
	return restok_define(name, owner);
	}

static int define(const struct configuration* cnf)
	{
	if (cnf == NULL)
		return -1;
	int success = 0;

	// If system pkg is present, tie it into restok default root.
	// The package providing _system_ will not show up in tokens.
	if (cnf->pkg_system)
		{
		cnf->pkg_system->id = register_resource(RESTOK_DEFAULT_ROOT, 0);
		if (cnf->pkg_system->alias)
			cnf->pkg_system->alias->id = cnf->pkg_system->id;
		}

	// Update kernel side token database
	for (struct package* pkg = cnf->pkg_first; pkg != NULL; pkg = pkg->next)
		{
		//warnx("Processing pkg '%s' id=%ld", pkg->name, pkg->id);
		if (pkg->alias != pkg)
			{
			// The package either "unfinished" or alias.
			if (pkg->alias)
				{
				//warnx("Skipping pkg '%s' as alias", pkg->name);
				continue;
				}
			// If package is not "finished", automaticly finish
			// it here.
			pkg->alias = pkg;
			// ..for now need to allow, fix later!
			//warnx("Ingoring unfinished package '%s'", pkg->name);
			//continue;
			}

		if (pkg->id == 0 && pkg->name[0] != '_' && (pkg->provide || pkg->objects))
			{
			/* Package is registered with kernel only if it
			   provides tokens or has objects, and name does
			   not start with '_'
			*/
			pkg->id = register_resource(pkg->name, 0);
			//warnx("Registered pkg '%s' with id=%ld", pkg->name, pkg->id);
			if (pkg->id <= 0)
				{
				warnx("Could not register package '%s' for the kernel module -- return id %d",
					pkg->name, (int)pkg->id);
				success = -1;
				pkg->id = -1;
				}
			}

		for (struct atom* a = pkg->provide; a != NULL; a = a->next)
			{
			creds_value_t value;
			creds_type_t type;

			if (a->id != CREDS_BAD)
				// Predefined resource id (for now: users and groups
				// from /etc/group and /etc/passwd)
				continue;
			if (pkg != cnf->pkg_domains && a->token[0] == '/')
				continue; // Ingore tokens for paths

			type = resolve_resource(pkg->name, a->token, value);
			//warnx("Resolving '%s::%s' [%d] id=%d::%ld", pkg->name, a->token, (int)a, type, value);

			if (type != CREDS_BAD)
				{
				if (pkg->creds_type == CREDS_BAD)
					{
					// If the type of resources defined by the package is unknown
					// at this point, this defines it.
					pkg->creds_type = type;
					}
				else if (pkg->creds_type != type)
					{
					// However, once defined, *ALL* resources must of the same type!
					warnx("Resource '%s::%s' has conflicting resource type %d != %d "
						"-- all resources of package must be of same type",
						pkg->name, a->token, type, pkg->creds_type);
					success = -1; // Fail the load!
					}
				a->id = value;
				continue;
				}

			if (pkg->creds_type == CREDS_BAD)
				pkg->creds_type = type_GRP; // Tokens are of type GRP!
			else if (pkg->creds_type != type_GRP)
				{
				warnx("Resource '%s::%s' cannot be defined -- package not configured for GRP (%d) but %d",
					pkg->name, a->token, type_GRP, pkg->creds_type);
				success = -1;
				continue;
				}

			if (pkg->id <= 0)
				{
				warnx("Resource '%s::%s' cannot be registered -- package is not registerable",
					pkg->name, a->token);
				continue;
				}
			a->id = register_resource(a->token, pkg->id);
			//warnx("Registered token '%s::%s' id=%ld", pkg->name, a->token, pkg->id);
			if (a->id == 0)
				continue;

			else if (a->id < 0)
				{
				warn("Could not register token '%s::%s' -- return id %d", pkg->name, a->token, (int)a->id);
				success = -1;
				continue;
				}
			//warnx("\tRegistered %s(%d):%s(%d)", pkg->name, (int)pkg->id, a->token, (int)a->id);
			}
		}
	return success;
	}

// Return the credential type code from credential struct
//
// If the direct provider does not have type, then the "master"
// pointed by alias defines it. (Note GRP/GID is an example where
// alias and master do have different type).
static creds_type_t credential_type(const struct credential &credential)
	{
	creds_type_t type = credential.provider->creds_type;
	if (type == CREDS_BAD && credential.provider->alias)
		type = credential.provider->alias->creds_type;
	return type;
	}


// Collect credentials for an object
// - cnf: The configuration (list of all packages)
// - policy_flags: The policy mode to be associated with the credentials set
// - creds: Returns the collected credentials set
// - pkg: The package owning the object
// - obj: The object self (executable)
//
// The final list of the credentials for the 'obj' includes all
// credentials requested for the object either by the owning package
// self, or other packages.
static int build_creds(const struct configuration& cnf, const int policy_flags,
	creds_t *creds, const struct package *pkg, const atom* obj,
	const creds_t *limit)
	{
	if (type_GRP == CREDS_BAD)
		return -1;

	// Always include the package token, if valid ("pkgname::pkgname")
	if (pkg->id > 0  && pkg != cnf.pkg_domains)
		{
		if (creds_add(creds, type_GRP, pkg->id))
			{
			warnx("Couldn't add package id for '%s' to policy", pkg->name);
			return -1;
			}
		}
	// Always include the package source token, if present ("SRC::pkgorigin")
	if (pkg->source && pkg->source->id != CREDS_BAD)
		{
		//warnx("\t* adding source token '%s' for '%s::%s'", pkg->source->token, pkg->name, obj->token);
		if (creds_add(creds, type_GRP, pkg->source->id))
			{
			warnx("Couldn't add source id for '%s' to policy", pkg->name);
			return 1;
			}
		}

	// Apply all modifications to the object from all packages:
	// for each package, search the object list to find whether
	// that package adds something to the credentials list of the
	// object being processed.
	//
	// This process also includes the case p==pkg: the credentials
	// requested by the object owning package are also included
	// and handled here.
	for (const struct package *p = cnf.pkg_first; p != NULL; p = p->next)
		for (const struct object *ref = p->objects; ref != NULL; ref = ref->next)
			{
			if (ref->path != obj)
				continue; // ... does not refer to current obj

			// Reference located, add the listed credentials
			for (struct credential* r = ref->request; r != NULL; r = r->next)
				{
				creds_type_t pkg_type = credential_type(*r);
				if (r->resource->id == CREDS_BAD || pkg_type == CREDS_BAD)
					{
					warnx("Policy for '%s' refers to undefined resource '%s(%d::%ld)::%s(%ld)' -- ignored",
						ref->path->token,
						r->provider->name, pkg_type, r->provider->id,
						r->resource->token, r->resource->id);
					continue;
					}
				if (limit && !creds_have_p(*limit, pkg_type, r->resource->id))
					continue;
				if (creds_add(creds, pkg_type, r->resource->id))
					{
					warnx("Couldn't add '%s:%s to policy", r->provider->name, r->resource->token);
					return -1;
					}
				}
			}
	return 0;
	}

// Generate and load the policies defined by the configuration
static int activate(const struct configuration* cnf)
	{
	if (cnf == NULL)
		return -1; // No memory or bad configuration...
	if (type_GRP == CREDS_BAD)
		return -1; // Kernel support faulty or missing.. no point of trying to activate!

	int result = 0;

	// The object "SRC::SRC" is defined to be holding the list of
	// credentials which are used in the origin checking
	// procedure. If credential is this set, and task is holding
	// it, then any code loaded into the address space, must come
	// from a origin that can also grant this credential. Construct
	// this list into 'src_creds'.
	const struct atom *src_obj = cnf->pkg_domains ? cnf->find_resource(cnf->pkg_domains, "SRC", 3) : NULL;
	creds_t src_creds = creds_init();
	if (src_obj)
		result = build_creds(*cnf, 0, &src_creds, cnf->pkg_domains, src_obj, NULL);
	else
		warnx("Didn't find object for 'SRC::SRC'");

	// Reuse the same credentials blob for everything -- the blob
	// grows to hold the largest credentials set below. Saves a
	// little, because creds_add does not need to allocate a new
	// blob for each list being constructed -- we just clear this.
	creds_t creds = creds_init();
	for (const struct package *pkg = cnf->pkg_first; pkg != NULL; pkg = pkg->next)
		{
		const creds_t *limit = NULL;

		if (pkg == cnf->pkg_socket)
			{
			const struct package *p = pkg->alias;
			if (!p)
				continue;
			//warnx("Processing socket package %s", p->name);

			struct skt_ioc_args *skt = socket_policy_start();
			if (!skt)
				{
				warnx("Socket policy setup failed");
				continue;
				}
			int errors = 0;
			for (const struct object *o = p->objects; o != NULL; o  = o->next)
				{
				if (o->path->token[0] != '/')
					continue;
				// +1 skips over the '/' in front of object name
				const char *socketstr = o->path->token+1;
				creds_type_t type = CREDS_BAD;
				creds_value_t value = CREDS_BAD;
				struct credential* r = o->request;
				if (r)
					{
					type = credential_type(*r);
					value = r->resource->id;
					//warnx("socket policy for '%s' requires '%s::%s' [%d::%d]",
					//	socketstr, r->provider->name, r->resource->token,
					//	(int)type, (int)value);
					if (r->next)
						warnx("Ignoring extra crentials for the object %s",
							socketstr);
					}
				else
					warnx("No credentials for the object %s", socketstr);
				if (socket_policy_add(skt, socketstr, type, value))
					{
					warnx("Failed adding socket policy for '%s'", socketstr);
					errors += 1;
					}
				}
			if (errors)
				{
				warnx("Not loading new socket policy due to errors");
				socket_policy_abort(skt);
				}
			else if (socket_policy_finish(skt))
				warnx("Failed loading new socket policy");
			continue;
			}
		else if (pkg == cnf->pkg_socket->alias)
			continue; // Don't load socket policy twice
		
		if (pkg == cnf->pkg_domains)
			{
			// Processing a "SRC::" objects.

			// Set the 'limit'. This removes all
			// credentials which are not part of the
			// origin checking set (no point in loading
			// such into kernel).
			limit = &src_creds;
			}
		
		// Process objects owned by this package
		for (struct atom *obj = pkg->provide; obj != NULL; obj = obj->next)
			{
			const char *prefix = pkg->name;

			int policy_flags = -1;

			if (pkg == cnf->pkg_domains)
				{
				// Building an "origin" policy list
				// (e.g. what credentials can this
				// origin grant). The policy flags
				// have no meaning, use 0 for now.
				policy_flags = 0;
				// BEWARE: for origin names,
				// obj->token[0] can be '/', but we
				// still need the prefix!
				}
			else
				{
				// Building a credential list for an
				// object defined by the package. The
				// object must also appear in the
				// objects of current package and
				// define the mode for the policy.
				for (const struct object *o = pkg->objects; o != NULL; o  = o->next)
					if (o->path == obj)
						{
						policy_flags = o->policy;
						break;
						}
				// Do not prefix plain executable paths
				// with package name.
				if (obj->token[0] == '/')
					prefix = NULL;
				}
			if (policy_flags == -1)
				continue;
			
			// Build new policy credentials vector
			creds_clear(creds);
			//warnx("-- New creds for '%s::%s", pkg->name, obj->token);
			result = build_creds(*cnf, policy_flags, &creds, pkg, obj, limit);
			if (result)
				goto out;

			if (load_object_creds(prefix, policy_flags, obj->token, creds))
				// For now, load error is not considered fatal...
				warnx("Failed loading policy for '%s::%s'", pkg->name, obj->token);
			}
		}
out:
	creds_free(creds);
	creds_free(src_creds);
	return result;
	}

static void dump_raw(const struct configuration* cnf)
	{
	if (cnf == NULL)
		return;
	cnf->db.Dump(printf);
	}

static void dump(const struct configuration* cnf)
	{
	if (cnf == NULL)
		return; // Nothing to print

	for (const struct package* pkg = cnf->pkg_first; pkg != NULL; pkg = pkg->next)
		{
		if (pkg->alias != pkg)
			continue; // alias package, or undefined packaged
		if (strcmp("GID", pkg->name) == 0 || strcmp("UID", pkg->name) == 0)
			continue; // skip generated packages.

		if (pkg == cnf->pkg_domains)
			continue; // Don't dump 'SRC::'

		printf("Package: %s\n", pkg != cnf->pkg_system ? pkg->name : system_pkg_name);
		if (pkg->source)
			printf("Source: %s\n", pkg->source->token);
		// Oh well, a bit brute force to get aliases out...
		for (const struct package* alias = cnf->pkg_first; alias != NULL; alias = alias->next)
			if (alias != pkg && alias->alias == pkg)
				printf("Provides:\n\t%s\n", alias != cnf->pkg_system ? alias->name : system_pkg_name);

		// List provide resources
		const char *tag = "Provide:\n";
		for (const struct atom* a = pkg->provide; a != NULL; a = a->next)
			// "provide" includes some file paths, don't "provide" them!
			if (a->token[0] != '/')
				{
				printf("%s\t%s\n", tag, a->token);
				tag = "";
				}

		for (const struct object* exe = pkg->objects; exe != NULL; exe = exe->next)
			{
			if (exe->registry != pkg)
				printf("Object: %s::%s\n", exe->registry->name, exe->path->token);
			else
				printf("Object: %s\n", exe->path->token);

			// Need to dump Policy, if not the default one
			int def_mask;
			int def_policy = credp_str2flags(NULL, &def_mask);
			if (exe->policy_mask && (exe->policy_mask != def_mask || exe->policy != def_policy))
				{
				const size_t size = credp_flags2str(exe->policy, exe->policy_mask, NULL, 0) + 1;
				char buf[size];
				credp_flags2str(exe->policy, exe->policy_mask, buf, size);
				printf("Policy: %s\n", buf);
				}
			if (exe->request)
				{
				printf("Request:\n");
				for (const struct credential* r = exe->request; r != NULL; r = r->next)
					{
					if (cnf->pkg_system == NULL ||
						(r->provider != cnf->pkg_system &&
						r->provider != cnf->pkg_system->alias))
						printf("\t%s::%s\n", r->provider->name, r->resource->token);
					else
						printf("\t%s\n", r->resource->token);
					}
				}
			for (struct plain_data* data = exe->data; data; data = data->next)
				printf("%s:%.*s\n", data->field->tag->name, data->size, data->data);
			}
		printf("\n");
		}
	}

static int be_manager(void)
	{
#ifndef RESTOK_RESOURCE_MANAGER
	// Define and switch the TCB locking to use 'tcb' token
	// This is NOOP, if the switch has already been made
	// The initial required permission is CAP_MAC_ADMIN.
	long root = register_resource(RESTOK_DEFAULT_ROOT, 0);
	if (root > 0)
		{
		// We have 'tcb', but it may be due to CAP_MAC_ADMIN,
		// do the switch.
		long tcb = register_resource("tcb", root);
		if (tcb < 0)
			{
			warn("Failed defining the 'tcb'");
			return tcb;
			}
		creds_value_t value;
		creds_type_t type = creds_str2creds("tcb", &value);
		if (type == CREDS_BAD)
			{
			warnx("Failed translation of 'tcb'");
			return -EINVAL;
			}
		// Before switch, make sure I've the token
		creds_t creds = creds_gettask(0);
		creds_add(&creds, type, value);
		creds_set(creds);
		creds_free(creds);
		if (restok_setlock(value) == 0)
			return 0; // Success
		}

// For a transition period, keep the "old way" of hacking tcb. This
// may become useful, if this library compiled with the new kernel
// headers gets installed to a device with older kernel.
//
// !! THIS IS TRANSITIONAL CODE WHICH WILL BE REMOVED LATER !!
//
#define RESTOK_RESOURCE_MANAGER "tcb"

	warnx("No tcb -- trying old way");
#endif
	/* Try to ensure that 'tcb' is defined -- ignore errors */
	register_resource(RESTOK_RESOURCE_MANAGER, restok_locate(RESTOK_DEFAULT_ROOT));

	/* Icky temp fix: grant self the right to manager the resources! */
	long manager = restok_locate(RESTOK_RESOURCE_MANAGER);
	if (manager > 0)
		{
		/* To avoid unnecessary error message from
		 * setgroups, find out if the current task is
		 * already the manager -- no setgroups needed.
		 */
		gid_t groups[32]; /* .. as temp fix. assuming 32 groups is enough */
		int count = getgroups(32, groups);
		/* if error, count is negative and following loop is NOOP */
		for (int i = 0; i < count; ++i)
			if (groups[i] == (gid_t)manager)
				return 0; /* Already manager */

		/* manager not found, try to add it */
		groups[0] = (gid_t)manager;
		if (setgroups(1, groups))
			warn("Failed setgroups for resource manager");
		else
			return 0; /* SUCCESS */
		}
	/* FAILED */
	warnx("Could not get 'tcb'");
	return -1;
	}

/* The plain C entry points */
__BEGIN_DECLS


const void* restok_cnf_load(const char* path, const char* update, size_t update_len)
	{
	return load_cnf(path ? path : "/var/lib/aegis/restok/restok.conf", update, update_len);
	}

int restok_cnf_define(const void* cnf)
	{
	return define((const RESTOK::configuration*)cnf);
	}

int restok_cnf_activate(const void* cnf)
	{
	return activate((const RESTOK::configuration*)cnf);
	}

void restok_cnf_dump(const void *cnf)
	{
	return dump((const RESTOK::configuration*)cnf);
	}

void restok_cnf_dump_raw(const void *cnf)
	{
	return dump_raw((const RESTOK::configuration*)cnf);
	}

void restok_cnf_free(const void* cnf)
	{
	configuration::release((RESTOK::configuration*)cnf);
	}

int restok_cnf_is_untrusted(const void* cnf)
	{
	return cnf ? ((const RESTOK::configuration*)cnf)->is_untrusted : 0;
	}

void restok_cnf_mark_untrusted(const void* cnf)
	{
	if (cnf)
		((RESTOK::configuration*)cnf)->is_untrusted = 1;
	}

int restok_be_manager()
	{
	return be_manager();
	}

__END_DECLS
