/*
 * This file is part of AEGIS
 *
 * Copyright (C) 2009 Nokia Corporation
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2.1 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 *
 * Author: Markku Savela
 */

#ifndef _LIBRESTOK_H_
#define _LIBRESTOK_H_

#include <features.h>

/// @cond IGNORE
__BEGIN_DECLS
/// @endcond

/*!
 * Load configuration and return a handle.
 *
 * Load the configuration and return a handle for
 * the in memory configuration. When not needed,
 * the configuration must be released by
 * restok_cnf_free.

 *
 * @param path Configuration file (or NULL, if default)
 * @param update Merge update file or buffer (or NULL, if no merge)
 * @param update_len Length of the update buffer (or zero, if from file).
 * @return Non-NULL handle, when load successful.

 */
extern const void* restok_cnf_load(const char *path, const char* update, size_t update_len);

/*!
 * Define all resources with the kernel module.
 *
 * Define all dynamic resources named in the current
 * configuration. This function is only avaible for
 * the component acting as a resource manager/launcer.
 *
 * @param[in] cnf The handle of the configuration
 * @return 0 on success, < 0 on error.
 *
 * @require resource-manager
 * @credential tcb Credential required to sucessfully define resource token.
 */
extern int restok_cnf_define(const void* cnf);

/*!
 * Activate the actual policies defined in the configuration
 *
 * This goes through the configuration and generates a
 * credentials setting policy for each defined executable.

 * @credential tcb Credential required to sucessfully activate configuration.
 */
extern int restok_cnf_activate(const void* cnf);

/*!
 * Dump textual presentation of the configuration
 */
extern void restok_cnf_dump(const void* cnf);
extern void restok_cnf_dump_raw(const void* cnf);

/*!
 * Release configuration handle.
 *
 * @param[in] cnf The configuration handle returned by restok_cnf_load.
 */
extern void restok_cnf_free(const void* cnf);

/*!
 * Return non-zero, if loaded configuration is untrusted
 *
 * @param[in] cnf The configuration handle returned by restok_cnf_load.
 *
 * The configuration is untrusted when any of the loaded is lacking a
 * signature or fails signature check.
 */
int restok_cnf_is_untrusted(const void* cnf);

/*!
 * Mark configuration untrusted
 *
 * @param[in] cnf The configuration handle returned by restok_cnf_load.
 *
 */
void restok_cnf_mark_untrusted(const void* cnf);

/*!
 * Attempt to become a resource manager.
 *
 * This function is temporary usage. In a properly
 * secured system, no process can self upgrade
 * itself into a resource manager.
 *
 * @return 0 on success, -1 on error
 * @credential CAP::CAP_MAC_ADMIN Credential required to use this function after the first switch
 * @credential tcb Credential required to use this function after the first switch
 */
extern int restok_be_manager();


__END_DECLS

#endif
