/*
 * This file is part of AEGIS
 *
 * Copyright (C) 2009-2010 Nokia Corporation
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2.1 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 *
 * Author: Markku Savela
 */
#include <stddef.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <limits.h>
#include <ctype.h>
#include <errno.h>
#include <err.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/creds.h>
#include <fcntl.h>
#include "restok_def.h"
#include "restok_db.h"
#include "restok_cnf.h"
#ifdef HAVE_LINUX_AEGIS_CREDP_H
#include <linux/aegis/credp.h>
#endif

#define STRING(s) s,sizeof(s)-1

namespace RESTOK
	{
	// List of internally known packages and their
	// mapping to credentials types. The following
	// identifiers (RESTOK_) may not be the same
	// ones that are used by the kernel, thus a
	// mapping table is required.
	enum
		{
		RESTOK_UID = 0,
		RESTOK_GID,
		RESTOK_GRP,
		RESTOK_CAP,
		RESTOK_MAX,
		};

	static const size_t PRF_LEN = 3;
	static struct
		{
		creds_type_t type;
		const char name[PRF_LEN+1];
		}
	map[] =
		// Be carefull to keep initializers in same order with above enum!!!
		// (I wish C++ accepted the "[index] = value" type initializer, like
		// Gnu C does!)
		{
		{CREDS_BAD, "UID"},
		{CREDS_BAD, "GID"},
		{CREDS_BAD, "GRP"},
		{CREDS_BAD, "CAP"},
		};

	struct alloc_chunk
		{
		struct alloc_chunk *next;
		size_t available;
		char data[];
		};


	struct atom* configuration::find_resource(const struct package* pkg, const char* name, size_t size)
		{
		// First, go to the primary package, if
		// this is an alias package.
		if (pkg->alias)
			pkg = pkg->alias;

		for (struct atom *r = pkg->provide; r != NULL; r = r->next)
			if (size == r->len && memcmp(r->token, name, size) == 0)
				return r;
		return NULL;
		}

	struct atom* configuration::init_resource(struct atom* r, size_t size, struct package& provider)
		{
		if (!r)
			return NULL;

		r->id = CREDS_BAD;

		// Append the new provided resource to the end of the
		// list. This done by "brutish force" to find the end
		// of the list at each insert -- should be fixed some
		// day, if "provide" lists are long. Current assumption
		// is the that packages rarely have provide list and
		// even then it only has one or few entries...
		struct atom **h = &provider.provide;
		while (*h)
			h = &(*h)->next;
		*h = r;
		r->next = NULL;
		//r->refcnt = 0; ?
		r->len = size;
		if (r->len != size)
			{
			// *UNLIKELY*
			// If the size does not fit the len, then name is too long
			// (the name has been fully copied into r->token and can
			// be used in error message, however).
			ERRX("Resource name '%s::%.*s' is too long", provider.name, size, r->token);
			return NULL;
			}
		//warnx("Created: '%s::%s", provider.name, r->token);
		return r;
		}

	// Note: Implicitly advances Scanner by size!
	static struct atom* make_resource(DB::Scanner &scanner, size_t size, struct package& provider)
		{
		struct atom* r = (atom *)scanner.Reserve(offsetof(atom, token), size);
		return configuration::init_resource(r, size, provider);
		}

	struct package* configuration::find_package(const char* name, size_t size) const
		{
		for (struct package* p = pkg_first; p != NULL; p = p->next)
			if (size == p->len && memcmp(p->name, name, size) == 0)
				return p;
		return NULL;
		}

	struct package* configuration::init_package(struct package *pkg, size_t size)
		{
		if (!pkg)
			return NULL;

		pkg->provide = NULL;
		pkg->objects = NULL;
		pkg->alias = NULL; // marks as yet undefined package!
		pkg->len = size;
		if (pkg->len != size)
			{
			// If the size does not fit the len, then name is too long
			// (the name has been fully copied into pkg->name and can
			// be used in error message, however).
			ERRX("Package name '%.*s' is too long", size, pkg->name);
			return NULL;
			}

		pkg->creds_type = CREDS_BAD; // Unknown at this stage.
		if (size == PRF_LEN) // All internal names have fixed length now!
			for (size_t i = 0; i < sizeof(map) / sizeof(map[0]); ++i)
				if (memcmp(pkg->name, map[i].name, PRF_LEN) == 0)
					{
					pkg->creds_type = map[i].type;
					// For now, internal packages are not assigned
					// special identifiers, prevent attempt to
					// generate it by setting id to -1.
					pkg->id = -1;
					break;
					}
		
		// Add new package the list.
		pkg->next = NULL;
		if (pkg_first == NULL)
			pkg_first = pkg;	// The first package
		else
			pkg_last->next = pkg;	// Subsequent package
		pkg_last = pkg;
		return pkg;
		}

	// *WARNING* Contrary to many other allocations, this comes from
	// heap. The memory must be released by "free".
	struct package* configuration::make_registry(const char* name)
		{
		const size_t len = strlen(name);
		const size_t size = offsetof(struct package, name) + len + 1;
		struct package* reg = (struct package*)malloc(size);
		if (reg)
			{
			memset(reg, 0, size);
			strcpy(reg->name, name);
			if (init_package(reg, len))
				reg->alias = reg;
			else
				{
				// init failed for some reason...
				free(reg);
				reg = NULL;
				}
			}
		return reg;
		}

	struct atom* configuration::register_object(DB::Scanner &scanner, size_t size, struct package& registry)
		{
		struct atom* r = find_resource(&registry, scanner.reading, size);
		if (r == NULL)
			{
			// The object has not been recorded yet, add now..
			r = make_resource(scanner, size, registry);
			//warnx("- new object %s::%s [len=%d]\n", registry.name, r->token, size);
			}
		else
			{
			// The has already been registered, return it.
			scanner.reading += size;
			//warnx("- old object %s::%s [len=%d]\n", registry.name, r->token, size);
			}
		return r;
		}
	
	// Note: Implicitly advances Scanner by size!
	struct package* configuration::make_package(DB::Scanner &scanner, size_t size)
		{
		struct package* pkg = (struct package *)scanner.Reserve(offsetof(struct package, name), size);
		return init_package(pkg, size);
		}

	// The next item in the input is assumed to be a resource reference. Check whether the
	// definition begins with a "name::" prefix. If not, return the default 'pkg' as
	// given in arguments. Otherwise process the prefix and pass over on input (adjusting
	// the size parameter and scanner.reading pointer).
	//
	// The name in prefix is a package name. If it has not been seen before, this creates
	// a new empty package with that name. Return the new or previously existing package.
	//
	// The return can be NULL only if the argument is NULL or new package creation failed.
	// If the passed pkg=NULL, detecting out of memory condition is not possible.
	struct package* configuration::assign_package(DB::Scanner &scanner, size_t& size, struct package* pkg)
		{
		char* ind = (char *)memchr(scanner.reading, ':', size);
		if (ind == NULL)
			return pkg; // No prefix, return the default pkg

		const size_t len = ind - scanner.reading;
		if (len+2 > size || ind[1] != ':')
			return pkg; // No prefix, return the default pkg

		// The resource starts with an explicit prefix, there is no way
		// we can return the default after that, thus 'pkg' is free to
		// be used locally... first, test if package is already known:

		pkg = find_package(scanner.reading, len);
		if (pkg)
			{
			scanner.reading += len + 2;
			size -= len + 2;
			}
		else if ((pkg = make_package(scanner, len)) != NULL)
			{
			// Skip over the trailing "::"
			scanner.reading += 2;
			// Need to update size for both name and '::'
			size -= len + 2;
			}
		return pkg;
		}

	// Cleanup up unclaimed unresolved resource references. Could just
	// abort on the first such occurrence, but it is nicer to get the
	// complete list at one parsing attempt -- thus this a bit more
	// complex solution.
	void configuration::clean_unresolved()
		{
		while (provide)
			{
			++unresolved;
			ERRX("Undefined resource '%s::%s' referenced, but not provided by the package",
				pkg_this ? pkg_this->name : "??", provide->token);
			provide = provide->next;
			}
		}

	void *configuration::alloc(size_t size, unsigned int mask)
		{
		// this function assumes the 'mask' is X consecutive least
		// significant bits (X >= 0), not checked in runtime!
		assert((mask & (mask + 1)) == 0);

		for (int done = 0; ;++done)
			{
			struct alloc_chunk *p;
			for (p = overflow; p != NULL; p = p->next)
				{
				if (size > p->available)
					continue; // No room in this chunk.
				const size_t spared = p->available - size;
				const size_t align = (size_t)(p->data+spared) & mask;
				if (align <= spared)
					{
					p->available = spared - align;
					DBG(warnx("Allocator -- biting %d bytes (aligned %d) from chunk, remaining = %d",
							size + align,
							align,
							p->available));
					return p->data + p->available;
					}
				// Ooops, align problem, which cannot
				// be corrected withing this chunk!
				}
			if (done)
				{
				ERRX("Allocator -- BAD CODING");
				return NULL;
				}
			// ...just make sure there is enough room, even
			// if align action hits...
			const size_t need = size + mask + 1;
			const size_t actual = need > 2000 ? need : 2000;
			p = (struct alloc_chunk*)malloc(offsetof(struct alloc_chunk, data) + actual);
			DBG(warnx("Allocator -- new chunk(%d) for size=%d %s", actual, size, p ? "OK" : "FAILED"));
			if (p == NULL)
				return NULL;
			p->available = actual;
			p->next = overflow;
			overflow = p;
			// Use the normal code to chew out a piece from
			// newly allocated chunk (should always succeed,
			// the 'done' bit is just a safeguard...)
			}
		}

	// Release everything allocated to the configuration and the configuration itself
	void configuration::release(struct configuration* cnf)
		{
		// Overflow chunks have been separately allocated.
		while (cnf->overflow)
			{
			struct alloc_chunk *tmp = cnf->overflow;
			cnf->overflow = tmp->next;
			free(tmp);
			}
		if (cnf->passwd)
			free(cnf->passwd);
		if (cnf->group)
			free(cnf->group);

		// The dynamically created registry packages are separate
		if (cnf->pkg_domains)
			free(cnf->pkg_domains);
		if (cnf->pkg_system)
			free(cnf->pkg_system);
		if (cnf->pkg_socket)
			free(cnf->pkg_socket);
		free(cnf);
		}


	// unused for now...
	static int PrintTokens(const struct credential* tokens, DB::PrintF printer)
		{
		int total = 0;
		while (tokens)
			{
			int count = printer("\n\t%s::%s", tokens->provider->name, tokens->resource->token);
			if (count < 0)
				return count;
			total += count;
			tokens = tokens->next;
			}
		return total;
		}

	// unused for now...
	static int PrintTokens(const struct atom* atoms, DB::PrintF printer)
		{
		int total = 0;
		while (atoms)
			{
			int count = printer("\n\t%s", atoms->token);
			if (count < 0)
				return count;
			total += count;
			atoms = atoms->next;
			}
		return total;
		}

	// This is just doing the shared work for the Package and Provides tags.
	// *Guaranteed* If ret == 0, then pkg != NULL
	int configuration::prepare_package(DB::Scanner &scanner, struct DB::field_value& field, int len, struct package*& pkg)
		{
		char *const end = scanner.reading + len;
		size_t tail;
		size_t size;

		(void)scanner.NextItem(end, '\n', size, tail);
		if (size <= 0 || scanner.reading + size + tail != end)
			{
			ERRX("The %s tag '%.*s' has bad syntax", field.tag->name, len, scanner.reading);
			return -2;
			}

		// The external 'name' of the RESTOK_DEFAULT_ROOT is
		// "_system_". Whenever this is referenced, it must be
		// substited with the internal real name.
		if (size == sizeof(system_pkg_name)-1 && memcmp(system_pkg_name, scanner.reading, size) == 0)
			{
			char *tmp = scanner.reading + size - sizeof(RESTOK_DEFAULT_ROOT) + 1;
			if (tmp < scanner.loading)
				{
				// NOTE: This should not happen, it is
				// totally unfair reason to reject the
				// configuration.
				// This will never happen if length of RESTOK_DEFAULT_ROOT is
				// less than system_pkg_name!
				ERRX("No room to replace '%s' with '" RESTOK_DEFAULT_ROOT "'", system_pkg_name);
				return -1;
				}
			// Replace in parsing buffer, update relevant pointers and vars.
			size = sizeof(RESTOK_DEFAULT_ROOT) - 1;
			memcpy(tmp, RESTOK_DEFAULT_ROOT, size);
			scanner.reading = tmp;
			len = end - scanner.reading;
			}
		
		// Check duplicate package
		pkg = find_package(scanner.reading, size);
		if (pkg == NULL)
			{
			pkg = make_package(scanner, size);
			if (pkg == NULL)
				return -1; // Out of memory
			return 0;
			}
		else if (pkg->alias)
			{
			if (!is_update)
				{
				ERRX("Duplicate package '%.*s'", len, scanner.reading);
				return -2;
				}
			
			// Fully defined package/alias is being redefined by a
			// update definition.

			// Need to nullify possible alias references...
			for (struct package* p = pkg_first; p != NULL; p = p->next)
				if (p->alias == pkg)
					p->alias = NULL;
			pkg->alias = NULL;
			}

		// Append the provide list from existing package to the
		// context. This list contains the forward references to
		// credentials that this package is expected to provide.
		// These must be provided by the package, or configuration is bad.
		while (pkg->provide)
			{
			struct atom* p = pkg->provide;
			pkg->provide = p->next;
			p->next = NULL;

			// Only need to consider tokens which are actually
			// referenced from outside the package.
			if (p->refcnt == 0)
				continue;
			//warnx("\t* processing forward ref for '%s::%s'", pkg->name, p->token);
			for (struct atom** h = &provide;;)
				{
				if (*h == NULL)
					{
					*h = p;
					break;
					}
				else if ((*h)->len == p->len &&	memcmp((*h)->token, p->token, p->len) == 0)
					{
					// Ignore duplicates (just forget the struct), but save the
					// external references count.
					(*h)->refcnt += p->refcnt;
					break;
					}
				h = &(*h)->next;
				}
			}
		// Throw away the objects list
		pkg->objects = NULL;
		// Package source is unknown until explicitly set.
		pkg->source = NULL;
		return 0;
		}

	static int ParsePackage(DB::Scanner &scanner, int len, struct DB::field_value& field)
		{
		struct configuration & cnf = *(struct configuration*)scanner.user_data;
		char *const end = scanner.reading + len;

		if (cnf.guard)
			{
			ERRX("The package tag '%.*s' not allowed here", len, scanner.reading);
			return -2;
			}
		
		// Cleanup after previous package, if any...
		cnf.clean_unresolved();	// ... provide == NULL after this!
		// Initialize package specific work space:
		cnf.request = NULL;	// ... no default request list
		cnf.object = NULL;	// ... no current object
		cnf.pkg_data = NULL;	// ... no data collecting
		if (field.next != NULL)
			{
			ERRX("The package tag '%.*s' was not the first in record", len, scanner.reading);
			return -2;
			}

		struct package* pkg;
		const int ret = cnf.prepare_package(scanner, field, len, pkg);
		if (ret)
			return ret;
		field.value_ptr = pkg;
		
		// Assign it to use as a real package by making alias
		// non-NULL and to point self.
		pkg->alias = pkg;

		cnf.pkg_this = pkg;
		scanner.reading = end;
		return 0;
		}

	static int ParseProvides(DB::Scanner &scanner, int len, struct DB::field_value& field)
		{
		struct configuration & cnf = *(struct configuration*)scanner.user_data;
		char *const end = scanner.reading + len;

		if (cnf.guard)
			{
			ERRX("The Provides tag '%.*s' not allowed here", len, scanner.reading);
			return -2;
			}

		if (cnf.pkg_this == NULL)
			{
			ERRX("No current package");
			return -2;
			}

		struct package* pkg;
		const int ret = cnf.prepare_package(scanner, field, len, pkg);
		if (ret)
			return ret;
		field.value_ptr = pkg;
		if (pkg == cnf.pkg_this)
			{
			// ..instead of this error, could possibly ignore the provides?
			ERRX("Package '%s' provides self -- this is not acceptable", pkg->name);
			return -2;
			}
		// Assign it to use as alias package by making alias
		// non-NULL and to point to real package (= current).
		pkg->alias = cnf.pkg_this;

		// Special kludge: if the creds_type of current package is
		// not yet fixed, but the alias has defined creds_type, then
		// assign that also to the main package. Mainly to get
		//	Package: something
		//	Provides: CAP
		// ...to work
		if (cnf.pkg_this->creds_type == CREDS_BAD && pkg->creds_type != CREDS_BAD)
			cnf.pkg_this->creds_type = pkg->creds_type;

		scanner.reading = end;
		return 0;
		}

	static int PrintPackage(const struct DB::field_value& field, DB::PrintF printer)
		{
		struct package &pkg = *(struct package *)field.value_ptr;
		return printer(" %s", pkg.name);
		}

	static struct DB::field_ops fops_package =
		{
		ParsePackage,
		PrintPackage,
		};

	static int Parse_etc(DB::Scanner &scanner, int len, struct DB::field_value& field)
		{
		struct configuration & cnf = *(struct configuration*)scanner.user_data;
		char *const end = scanner.reading + len;

		// Generate synthetic resources from users and groups.

		if (cnf.pkg_this == NULL)
			{
			ERRX("No current package");
			return -2;
			}

		size_t tail;
		size_t size;
		(void)scanner.NextItem(end, ':', size, tail);

		if (cnf.pkg_this->creds_type != map[RESTOK_GID].type &&
			cnf.pkg_this->creds_type != map[RESTOK_UID].type)
			{
			ERRX("Ignoring unknown tag '%.*s' while processing '%s'", size, scanner.reading, cnf.pkg_this->name);
			scanner.reading = end;
			field.value_ptr = NULL;
			return 0;
			}

		struct atom* r = make_resource(scanner, size, *cnf.pkg_this);
		if (r == NULL)
			return -1;
		field.value_int = 0;
		scanner.reading += tail;
		(void)scanner.NextItem(end, ':', size, tail); // Skip the passwd field.
		scanner.reading += size + tail;
		(void)scanner.NextItem(end, ':', size, tail); // Extract User or Group number.
		r->id = atoi(scanner.reading);
		DBGG(warnx("new resource: %ld (%s)", r->id, r->token));
		scanner.reading = end;
		return 0;
		}

	static int Print_etc(const struct DB::field_value& field, DB::PrintF printer)
		{
		// Do not print etc
		return 0;
		}

	static struct DB::field_ops fops_etc =
		{
		Parse_etc,
		Print_etc,
		};


	static int PrintTokens(const struct DB::field_value& field, DB::PrintF printer);
	static int ParseTokens(DB::Scanner &scanner, int len, struct DB::field_value& field);
	static struct DB::field_ops fops_tokens =
		{
		ParseTokens,
		PrintTokens,
		};

	static int PrintObject(const struct DB::field_value& field, DB::PrintF printer);
	static int ParseObject(DB::Scanner &scanner, int len, struct DB::field_value& field);
	static struct DB::field_ops fops_object =
		{
		ParseObject,
		PrintObject,
		};


	static struct DB::field_ops fops_provides =
		{
		ParseProvides,
		PrintPackage,
		};


	static int ParseData(DB::Scanner &scanner, int len, struct DB::field_value& field)
		{
		struct configuration & cnf = *(struct configuration*)scanner.user_data;
		char *const end = scanner.reading + len;
		struct plain_data* data = NULL;

		if (cnf.pkg_this == NULL)
			{
			ERRX("No current package");
			return -2;
			}
		if (cnf.pkg_data)
			{
			data = (struct plain_data *)scanner.Reserve(offsetof(plain_data, data), len);
			if (data == NULL)
				{
				ERRX("Out of memory while reading data '%s' for '%s'", field.tag->name, cnf.pkg_this->name);
				return -1;
				}
			data->next = NULL;
			data->field = &field;
			data->size = len;
			*cnf.pkg_data = data;
			cnf.pkg_data = &data->next;
			}
		field.value_ptr = data; // Mark field as processed.
		scanner.reading = end;
		return 0;
		}

	static int PrintData(const struct DB::field_value& field, DB::PrintF printer)
		{
		struct plain_data* data = (struct plain_data*)field.value_ptr;
		if (data)
			return printer(" %s", data->data);
		return 0;
		}

	static struct DB::field_ops fops_data =
		{
		ParseData,
		PrintData,
		};

	static int ParseSource(DB::Scanner &scanner, int len, struct DB::field_value& field)
		{
		struct configuration & cnf = *(struct configuration*)scanner.user_data;
		char *const end = scanner.reading + len;

		if (cnf.pkg_this == NULL)
			{
			ERRX("No current package");
			return -2;
			}

		size_t tail;
		size_t size;
		(void)scanner.NextItem(end, '\n', size, tail);
		if (size <= 0 || scanner.reading + size + tail != end)
			{
			ERRX("The source tag '%.*s' has bad syntax", len, scanner.reading);
			return -2;
			}
		if (cnf.pkg_this->source)
			{
			ERRX("Package '%s' has multiple source definitions: '%s' and '%.*s'",
				cnf.pkg_this->name,
				cnf.pkg_this->source->token,
				size, scanner.reading);
			return -2;
			}
		struct package *registry = cnf.assign_package(scanner, size, cnf.pkg_domains);
		if (registry != cnf.pkg_domains)
			{
			ERRX("Bad source %s::%.*s in package '%s'",
				registry->name,
				size, scanner.reading,
				cnf.pkg_this->name);
			return -2;
			}
		cnf.pkg_this->source = cnf.register_object(scanner, size, *registry);
		if (cnf.pkg_this->source == NULL)
			{
			ERRX("Out of memory in parsing package '%s' source '%.*s'",
				cnf.pkg_this->name, size, scanner.reading);
			return -1;
			}
		field.value_ptr = cnf.pkg_this->source;
		scanner.reading = end;
		return 0;
		}

	static int PrintSource(const struct DB::field_value& field, DB::PrintF printer)
		{
		struct atom &atom = *(struct atom *)field.value_ptr;
		return printer(" %s", atom.token);
		}

	static struct DB::field_ops fops_source =
		{
		ParseSource,
		PrintSource,
		};

	static int ParsePolicy(DB::Scanner &scanner, int len, struct DB::field_value& field)
		{
		struct configuration & cnf = *(struct configuration*)scanner.user_data;
		const char *pkg_name = cnf.pkg_this ? cnf.pkg_this->name : "[no-package]";
		if (cnf.object == NULL)
			{
			warnx("'Policy: %.*s' without preceding Object in package '%s'\n",
				len, scanner.reading, pkg_name);
			return -2;
			}

		const char *str = (const char *)scanner.Reserve(0, len);
		if (!str)
			return -1; /* Out of memory! */
		field.value_ptr = (void *)str;
		cnf.object->policy = credp_str2flags(str, &cnf.object->policy_mask);
		//warnx("Got policy: '%s' -> flags=%d mask=%d\n", str, cnf.object->policy, cnf.object->policy_mask);
		if (cnf.object->policy == -1)
			{
			if (cnf.object->policy_mask)
				warnx("'Polixy: %s' has conflicting keyword(s) in package '%s'\n",
					str, pkg_name);
			else
				warnx("'Policy: %s' has unknown keyword(s) in package '%s'\n\n",
					str, pkg_name);
			//return -2;
			// For for now, let above errors pass
			}
		return 0;
		}
	
	static int PrintPolicy(const struct DB::field_value& field, DB::PrintF printer)
		{
		const char *str = (const char *)field.value_ptr;
		return printer("%s", str);
		}

	static struct DB::field_ops fops_policy =
		{
		ParsePolicy,
		PrintPolicy,
		};

	static int ParseGuard(DB::Scanner &scanner, int len, struct DB::field_value& field)
		{
		struct configuration & cnf = *(struct configuration*)scanner.user_data;
		field.value_int = 0; // Mark field as processed.
		cnf.guard = 1;
		scanner.reading = scanner.reading + len;
		return 0;
		}

	static int PrintGuard(const struct DB::field_value& field, DB::PrintF printer)
		{
		return 0;
		}

	static struct DB::field_ops fops_guard =
		{
		ParseGuard,
		PrintGuard,
		};

	DB_ATOM_STATIC(object,		&fops_object,	"Object");
	DB_ATOM_STATIC(request,		&fops_tokens,	"Request");
	DB_ATOM_STATIC(provide,		&fops_tokens,	"Provide");
	DB_ATOM_STATIC(package,		&fops_package,	"Package");
	DB_ATOM_STATIC(source,		&fops_source,	"Source");
	DB_ATOM_STATIC(provides,	&fops_provides, "Provides");
	DB_ATOM_STATIC(key,		&fops_data,	"Key");
	DB_ATOM_STATIC(rank,		&fops_data,	"Rank");
	DB_ATOM_STATIC(untrusted,	&fops_data,	"Untrusted");
	DB_ATOM_STATIC(recover,		&fops_data,	"Recover");
	DB_ATOM_STATIC(grant,		&fops_data,	"Grant");
	DB_ATOM_STATIC(policy,		&fops_policy,	"Policy");

	static const struct DB::atom_string* atom_list[] =
		{
		(DB::atom_string*)&package,
		(DB::atom_string*)&source,
		(DB::atom_string*)&provides,
		(DB::atom_string*)&request,
		(DB::atom_string*)&provide,
		(DB::atom_string*)&object,
		(DB::atom_string*)&key,
		(DB::atom_string*)&rank,
		(DB::atom_string*)&untrusted,
		(DB::atom_string*)&recover,
		(DB::atom_string*)&grant,
		(DB::atom_string*)&policy,
		NULL,
		};

	// This list is for parsing /etc/group and
	// /etc/passwd. Because they are not fully secured, accept
	// only a very limited set of tags while parsing them. Also,
	// use special "tags", which cannot collide with valid group
	// or user name.
	DB_ATOM_STATIC(etc_terminal,	&fops_etc,	"");
	DB_ATOM_STATIC(etc_guard,	&fops_guard,	"-");
	DB_ATOM_STATIC(etc_provides,	&fops_provides,	"-Provides");
	DB_ATOM_STATIC(etc_package,	&fops_package,	"-Package");

	static const struct DB::atom_string* etc_list[] =
		{
		(DB::atom_string*)&etc_package,
		(DB::atom_string*)&etc_provides,
		(DB::atom_string*)&etc_guard,
		(DB::atom_string*)&etc_terminal,
		NULL,
		};


	class RequestCb : public DB::SplitCb
		{
	public:
		RequestCb(struct configuration &cnf, struct credential **last) :
			SplitCb(), cnf(cnf), last(last) {}
		int Item(DB::Scanner& scanner, size_t len);

		struct configuration & cnf;
		struct credential **last;
		};


	int RequestCb::Item(DB::Scanner& scanner, size_t size)
		{
		if (size == 0)
			{
			ERRX("Request for empty credential string ignored in %s",
				cnf.pkg_this ? cnf.pkg_this->name : "(unknown)");
			// This is on borderline of syntax error, but for now let it pass
			return 0;
			}
		// If the "token" begins with ':', the this is some control flag
		// which need special handling. By default, ignore all such
		// silently.
		if (*scanner.reading == ':')
			{
			scanner.reading += size;
			return 0;
			}
		
		// Requesting a resource (for unqualified names, default first to
		// the resources provided locally by the current package: cnf.pkg_this
		// -- this assumes the "Provide" tag preceeds the "Request" tag!)
		struct package* pkg  = cnf.assign_package(scanner, size, cnf.pkg_this);
		if (pkg == NULL)
			{
			ERRX("Cannot locate/make resource defining package for '%.*s'", size, scanner.reading);
			return -1;
			}
		// This resource should be found from the already known resources!
		struct atom* r = configuration::find_resource(pkg, scanner.reading, size);

		if (r == NULL)
			{
			if (pkg == cnf.pkg_this)
				{
				// If the unqualified resource was not provided by the
				// current package, assume the resource is/will be defined
				// by the system package.
				pkg = cnf.pkg_system;
				r = configuration::find_resource(pkg, scanner.reading, size);
				}
			if (r == NULL)
				{
				if (pkg->alias) // If package fully defined ...
					{
					ERRX("The package '%s' does not provide for resource '%.*s'", pkg->name, size, scanner.reading);
					return -2;
					}
				DBG(warnx("Possible forward request for a resource '%s::%.*s'", pkg->name, size, scanner.reading));
				// Try to resolve these later in the post processing phase. For now, just
				// add the resource to yet undefined package...
				r = make_resource(scanner, size, *pkg);
				if (r == NULL)
					return -1; // out of memory
				}
			else
				{
				// The space used by token string can be reclaimed
				// (because not needed for a new resource definition,
				// as in above)
				scanner.reading += size;
				}
			}
		else
			{
			// The space used by token string can be reclaimed
			// (because not needed for a new resource definition,
			// as in above)
			scanner.reading += size;
			}

		struct credential* t = (struct credential*)scanner.ReserveAligned(sizeof(credential));
		if (t == NULL)
			return -1; // out of memory
		t->next = NULL;
		t->resource = r;
		t->provider = pkg;
		r->refcnt += (pkg != cnf.pkg_this);

		// Append t to the 'request' list of the current
		// package or exectable.
		*last = t;
		last = &t->next;
		return 0;
		}
		
	class ProvideCb : public DB::SplitCb
		{
	public:
		ProvideCb(struct configuration &cnf) : SplitCb(), cnf(cnf) {}
		int Item(DB::Scanner& scanner, size_t size);

		struct configuration & cnf;
		};

	int ProvideCb::Item(DB::Scanner& scanner, size_t size)
		{
		if (size == 0)
			{
			ERRX("Provide for empty credential string ignored in %s",
				cnf.pkg_this ? cnf.pkg_this->name : "(unknown)");
			// This is on borderline of syntax error, but for now let it pass
			return 0;
			}

		// Providing a resource -- all resource names should be unqualified.
		struct package* pkg  = cnf.assign_package(scanner, size, cnf.pkg_this);
		if (pkg == NULL)
			{
			ERRX("Out of memory for '%.*s", size, scanner.reading);
			return -1;
			}
		if (pkg != cnf.pkg_this)
			{
			ERRX("Package '%s' tries to provide a credential '%s::%.*s' -- ignored",
				cnf.pkg_this->name, pkg->name, size, scanner.reading);

			//return -2;

			//'dpkg' wrapper accidentally let this one
			// pass from manifest, for now just ignore
			// with warning. Possibly return back to
			// original strict error in future.
			scanner.reading += size;
			return 0;
			}

		// Resolve pending forward references (saved in cnf.provide list).
		struct atom* r;
		for (struct atom** h = &cnf.provide; (r = *h) != NULL; h = &r->next)
			{
			if (r->len == size && memcmp(r->token, scanner.reading, size) == 0)
				{
				// Matched an existing forward reference, reuse that as a real
				// provided resource entry (move to real list).
				*h = r->next;

				(void)configuration::init_resource(r, size, *pkg);
				scanner.reading += size;
				return 0;
				}
			}
		// Not found in forward

		r = cnf.register_object(scanner, size, *pkg);
		if (r == NULL)
			return -1; // out of memory
		return 0;
		}

	static int ParseTokens(DB::Scanner &scanner, int len, struct DB::field_value& field)
		{
		char *const start = scanner.reading;
		struct configuration & cnf = *(struct configuration*)scanner.user_data;

		if (cnf.pkg_this == NULL)
			{
			ERRX("No current package");
			return -2;
			}

		if (field.tag == (struct DB::atom_string *)&request)
			{
			struct credential **head = cnf.object ? &cnf.object->request: &cnf.request;
			*head = NULL; // New request starts from empty list

			RequestCb cb(cnf, head);
			// split by any white space (isspace)
			int ret = scanner.SplitBy(len, 0, cb);
			field.value_ptr = *head;
			return ret;
			}
		else if (field.tag == (struct DB::atom_string *)&provide)
			{
			ProvideCb cb(cnf);
			// split by any white space (isspace)
			int ret = scanner.SplitBy(len, 0, cb);
			field.value_ptr = cnf.pkg_this->provide;
			return ret;
			}
		ERRX("Invalid tag '%s:%.*s'", field.tag->name, len, start);
		return -2;
		}

	static int PrintTokens(const struct DB::field_value& field, DB::PrintF printer)
		{
		if (field.tag == (struct DB::atom_string *)&request)
			{
			return PrintTokens((struct credential *)field.value_ptr, printer);
			}
		else if (field.tag == (struct DB::atom_string *)&provide)
			{
			return PrintTokens((struct atom *)field.value_ptr, printer);
			}
		return printer("<tokens printer>");
		}


	static int ParseObject(DB::Scanner &scanner, int len, struct DB::field_value& field)
		{
		struct configuration & cnf = *(struct configuration*)scanner.user_data;
		char *const end = scanner.reading + len;
		if (cnf.pkg_this == NULL)
			{
			ERRX("No current package");
			return -2;
			}

		size_t tail;
		size_t size;
		(void)scanner.NextItem(end, '\n', size, tail);
		if (size <= 0 || scanner.reading + size + tail != end)
			{
			ERRX("The object tag '%.*s' has bad syntax", len, scanner.reading);
			return -2;
			}

		struct package* registry = cnf.assign_package(scanner, size, cnf.pkg_this);
		if (registry == NULL)
			{
			ERRX("Object registry not properly initialized");
			return -2;
			}

		struct object **h, *e;
		for (h = &cnf.pkg_this->objects; (e = *h) != NULL; h = &e->next)
			{
			if (size == e->path->len &&
				e->registry == registry &&
				memcmp(e->path->token, scanner.reading, size) == 0)
				{
				// Object path has already been recorded in this package!
				ERRX("Duplicate definition of object '%s::%s'", e->registry->name, e->path->token);
				return -2;
				}
			}
		field.value_ptr = scanner.ReserveAligned(sizeof(struct object));
		e = (struct object *)field.value_ptr;
		if (e == NULL)
			return -1; // Out of memory
		*h = e;
		e->next = NULL;
		e->request = cnf.request; // Default to package list of tokens.
		e->registry = registry;

		struct atom* r = NULL;
		if (registry == cnf.pkg_this)
			{
			// Defining object in current package, process possible
			// unresolved references to this object.

			// Get the default policy and mask.
			e->policy = credp_str2flags(NULL, &e->policy_mask);
			for (struct atom** h = &cnf.provide; (r = *h) != NULL; h = &r->next)
				{
				if (r->len == size && memcmp(r->token, scanner.reading, size) == 0)
					{
					// Matched an existing forward reference, reuse that as a real
					// provided resource entry (move to real list).
					*h = r->next;
					
					(void)configuration::init_resource(r, size, *registry);
					scanner.reading += size;
					break;
					}
				}
			}
		if (r == NULL)
			{
			r = cnf.register_object(scanner, size, *registry);
			if (r == NULL)
				return -1;
			}
		r->refcnt += (registry != cnf.pkg_this);
		e->path = r;
		cnf.object = e;
		cnf.pkg_data = &e->data;
		scanner.reading = end;
		return 0;
		}

	static int PrintObject(const struct DB::field_value& field, DB::PrintF printer)
		{
		struct object& exe = *(struct object *)field.value_ptr;
		return printer(" %s", exe.path->token);
		}

	class Allocator : public DB::Overflow
		{
	public:

		Allocator(configuration& cnf) : cnf(cnf) {}
		void *ReserveAligned(size_t size, unsigned int mask);
	private:
		configuration& cnf;
		};

	void *Allocator::ReserveAligned(size_t size, unsigned int mask)
		{
		return cnf.alloc(size, mask);
		}


	void *load_etc(struct configuration *cnf, const char *prefix, const char *path,
		Allocator &overflow)
		{
		struct etc
			{
			struct DB::db_header db;
			char spare[100];
			char data[];
			};
		
		unsigned int flags;
		size_t size;
		struct etc *etc;

		etc = (struct etc *)DB::load_files(offsetof(struct etc, data),
			0, flags, size, prefix, path, NULL);
		if (!etc)
			{
			ERRX("Failed loading '%s'", path);
			return NULL;
			}
		
		// Parse the file
		cnf->guard = 0;
		int len = etc->db.Load(
			size - offsetof(struct etc, data),
			etc->data - etc->db.base,
			etc_list, cnf,
			NULL,
			overflow);
		cnf->guard = 0;
		
		if (len < 0)
			{
			free(etc);
			ERRX("Failed parsing '%s'", path);
			return NULL;
			}
		return etc;
		}
	
	const struct configuration* configuration::load(const char* path, const char* update, size_t update_len)
		{
		// If the "update-len" is ZERO and update is non-NULL, then
		// update is interpreted as filename. Otherwise, update
		// is a memory buffer containing the update.
		const char *update_file = update_len ? NULL : update;

		size_t size;
		unsigned int flags;

		map[RESTOK_UID].type = creds_str2creds("UID::", NULL);
		map[RESTOK_GID].type = creds_str2creds("GID::", NULL);
		map[RESTOK_GRP].type = creds_str2creds("GRP::", NULL);
		map[RESTOK_CAP].type = creds_str2creds("CAP::", NULL);

		configuration *cnf = (configuration *)DB::load_files
			(offsetof(struct configuration, data), update_len, flags, size,
			"", path,
			"\n", update_file,
			NULL);

		if (cnf == NULL)
			{
			ERRX("Allocating buffer size=%d failed", size);
			return NULL;
			}

		if (flags)
			{
			if (flags & ~DB::LOAD_NO_SIGNATURE)
				{
				ERRX("Could not get all requested files");
				free(cnf);
				return NULL;
				}
			cnf->is_untrusted = 1;
			// ...make this silent for now...
			//ERRX("Warning: one or more of files '/etc/passwd', '/etc/group' or '%s' was not signed", path);
			}

		cnf->is_update = (update != NULL);
		if (update_len && update)
			{
			// Update has been specified as memory buffer, copy it
			// in before parsing.
			memcpy(((char *)cnf) + size - update_len, update, update_len);
			}

		Allocator overflow(*cnf);
		cnf->pkg_domains = cnf->make_registry("SRC");
		cnf->pkg_socket = cnf->make_registry("_socket_");
		cnf->pkg_socket->alias = NULL;
		cnf->pkg_system = cnf->make_registry(RESTOK_DEFAULT_ROOT);
		cnf->pkg_system->alias = NULL;

		// Load and processs /etc/passwd and /etc/group (note:
		// the "-:" line prevents the recognition of
		// "-Package" and "-Provides" in the following data
		// (see cnf->guard).
		cnf->passwd = load_etc(cnf, "-Package: UID\n-:\n", "/etc/passwd", overflow);
		cnf->group = load_etc(cnf, "-Package: GID\n-Provides: GRP\n-:\n", "/etc/group", overflow);

		int len = cnf->db.Load(
			size - offsetof(struct configuration, data),
			cnf->data - cnf->db.base,
			atom_list, cnf,
			NULL,
			overflow);
		if (len < 0)
			{
			ERRX("Error in processing the policy");
			free(cnf);
			return NULL;
			}
		DBG(warnx("Load returned %d\n", len));
		cnf->clean_unresolved(); // ...for the last package, if any.

		if (cnf->unresolved)
			ERRX("Warning: configuration has %d unresolved resources", cnf->unresolved);
		DBGG(
			printf("\n----------------------\n"); 
			cnf->db.Dump(printf);
			printf("----------------------\n")
			);
		return cnf;
		}
	}
