/*
 * This file is part of AEGIS
 *
 * Copyright (C) 2009-2010 Nokia Corporation
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 *
 * Author: Markku Savela
 */

#ifndef RESTOK_CNF_H
#define RESTOK_CNF_H

extern "C"
	{
#ifdef HAVE_LINUX_AEGIS_RESTOK_H
#	include <linux/aegis/restok.h>
#else
#	include "restok_fallback.h"
#endif
	}

#include <sys/creds.h>
#include "restok_db.h"
#include "librestok.h"

// External name of the package which provides the default tokens
static const char system_pkg_name[] = "_system_";

namespace RESTOK
	{
	// Basic description of string token
	struct atom
		{
		// The tokens are linked together under provider
		// of the package.
		struct atom* next;
		// Count of external references to this token.
		// (number of references from other packages
		// than the providing package)
		unsigned int refcnt;
		// If the string corresponds to a credentials symbol,
		// the translated credentials value corresponding
		// the string is stored here. Value is valid, if not
		// CREDS_BAD. Initial value is CREDS_BAD.
		creds_value_t id;

		// The length of the token (max 65535 characters)
		// [need longer than 255, due to "tokenizing" paths sometime]
		unsigned short len;
		// ... followed by NUL-terminated token name string.
		// The name is "plain", without any "foo::" prefixes. When
		// used, the token names are always associated with
		// a package ("package::token"). If a reference omits
		// the package part, a special package "_system_::" is
		// the default.
		char token[];
		};

	// Credential reference
	struct credential
		{
		struct credential* next;	// Links a collection of credentials
		struct package* provider;	// Provider of the credential
		struct atom* resource;		// Name of the credential
		};

	// A structure to carry plain configuration data
	// (a data which needs to be kept around, but not
	// used by current program...
	struct plain_data
		{
		struct plain_data* next; 	// chaining of data
		const struct DB::field_value* field;	// containing field
		size_t size;			// The length of the data to follow
		char data[];			// The data
		};

	// Object reference describes credentials applied to an object
	// Current object classess:
	// - executable file
	// - package domain
	struct object
		{
		struct object* next;		// Links a collection of object references
		int policy;			// Policy modifiers assocated with the request
		int policy_mask;		// ... (indicate what bits are used in above).
		struct credential* request;	// The collection of requested credentials
		struct package *registry;	// The object owner
		struct atom* path;		// The object identifier
		struct plain_data* data;	// Other data associated with the object
		};

	// Describe a package
	struct package
		{
		struct package* next;

		// The generic meaning
		// alias = NULL,
		//	The undefined initial value.
		//	Structure does describe any real
		//	package yet (no "Package:" line seen)
		// alias = self,
		//	This is a full description of a package
		//	(the "Package:" line has been seen).
		// alias != self
		//	This is alias name for a package. The
		//	only field used is the creds_type (e.g.
		//	when package "grp" is aliased to "gid",
		//	grp:: gives CREDS_GRP, while gid:: gives
		//	CREDS_GID.
		struct package* alias;

		// The source of the package. The registry (provider)
		// is implicitly known (it is the pgk_domains package).
		struct atom* source;

		// The tokens provided (exported) by the package. This
		// includes all "symbols" that can be referenced from
		// other packages (currently: credential tokens and
		// file paths; pkg_domains, it's the domains).
		struct atom* provide;

		// The list of objects modified by this package. Includes
		// objects provided by this package and possibly others.
		struct object* objects;

		// The generic meaning of the id's
		// id = 0, the initial undefined value
		// id > 0, the package has been assigned an id
		// id < 0, the package does not use id, or assignment failed.
		long id;
		// The type of credentials values from this pkg
		creds_type_t creds_type;
		// The length of the name (max 255 characters)
		unsigned char len;
		// ...followed by NUL-terminated name of the package.
		char name[];
		};

	struct alloc_chunk;
	struct configuration
		{
		static void release(struct configuration *cnf);
		static struct atom* find_resource(const struct package* pkg, const char* name, size_t size);
		static struct atom* init_resource(struct atom* atom, size_t size, struct package& provider);

		//struct atom* find_resource(const char* name, size_t size, struct package*& provider) const;
		struct package* find_package(const char* name, size_t size) const;
		struct package* init_package(struct package* pkg, size_t size);
		struct package* make_package(DB::Scanner &scanner, size_t size);
		struct package* make_registry(const char* name);
		struct package* assign_package(DB::Scanner &scanner, size_t& size, struct package* pkg);
		struct atom* register_object(DB::Scanner &scanner, size_t size, struct package& registry);
		int prepare_package(DB::Scanner &scanner, struct DB::field_value& field, int len, struct package*& pkg);
		void clean_unresolved();
		static const struct configuration* load(const char* path, const char* update, size_t update_len);
		void *alloc(size_t size, unsigned int mask = 3);
		struct package *pkg_first;	// First package in the package list
		struct package *pkg_last;	// Last package in the package list
		struct package *pkg_system;	// The default system package.
		struct package *pkg_domains;	// Registry of known domains
		struct package *pkg_socket;	// Registry of socket policies

		// ...the following are transient working space used
		// during the parsing/loading (not valid after load)

		struct package *pkg_this;	// The current package being processed.
		struct object *object;		// The current object within package (or NULL).
		struct plain_data** pkg_data;	// End of the plain data chain.
		struct credential* request;	// The package default resource request list.
		struct atom* provide;		// The package forward resource provide list.

		void *passwd; // memory blob for /etc/passwd information
		void *group;  // memory blob for /etc/group information
		
		// ...data valid during and after the load

		unsigned int guard:1;		// 1 = "guard" seen (only in etc reading)
		unsigned int is_update:1;	// 1 = if load is for merged update.
		unsigned int is_untrusted:1;	// 1 = if loaded files failed signatures
		unsigned int unresolved;	// The count of unresolved resource references.
		struct alloc_chunk* overflow;	// Additional/overflow memory allocations.
		struct DB::db_header db;
		char spare[2000];		// Initial spare room for parsing.
		char data[];			// Start of the data
		};
	}
#endif
