/*
 * This file is part of AEGIS
 *
 * Copyright (C) 2009 Nokia Corporation
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2.1 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 *
 * Author: Markku Savela
 */
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <limits.h>
#include <err.h>
#include <ctype.h>
#include "restok_def.h"
#include "restok_db.h"

namespace RESTOK
	{
	namespace DB
		{

		// Only support 4 or 8 alignment.
		static const unsigned int align_mask = sizeof(void *) > 4 ? 7 : 3;

		//static const size_t atom_size = sizeof(struct atom_string);
		static const size_t field_size = sizeof(struct field_value);


		static int Parse_StringValue(Scanner &scanner, int len, struct field_value& field)
			{
			field.value_ptr = scanner.Reserve(0, len);
			return field.value_ptr ? 0 : -1;
			}

		static int Print_StringValue(const struct field_value& field, PrintF printer)
			{
			return printer("%s", (char *)field.value_ptr);
			};

		static struct field_ops fops_default =
			{
			Parse_StringValue,
			Print_StringValue
			};

		int db_header::Load(int size, int offset, const struct atom_string *atoms[], void *user_data, Body body_parser, Overflow& overflow)
			{
			Scanner scanner(*this, size, offset, user_data, overflow);
			return scanner.Load(atoms, body_parser);
			}

		int db_header::Dump(PrintF printer) const
			{
			for (const struct record_value *r = records; r != NULL; r = r->next)
				{
				for (const struct field_value* f = r->fields; f != NULL; f = f->next)
					{
					// Interpret field value == 0, as "do not dump this".
					if (f->value_int == 0)
						continue;
					
					if (printer("%s:", f->tag->name) < 0)
						return -1;
					if (f->tag->fops && f->tag->fops->print)
						if (f->tag->fops->print(*f, printer) < 0)
							return -1;
					if (printer("\n") < 0)
						return -1;
					}
				if (printer("\n") < 0)
					return -1;
				}
			return 0;
			}

		int Scanner::Load(const struct atom_string *atoms[], Body body_parser)
			{
			struct record_value **last_record = &db.records;
			struct field_value **last_field = NULL;
			struct field_value *first_field = NULL;

			db.records = NULL;
			db.atoms = NULL;
			db.atoms_static = atoms;

			while (reading < ending)
				{
				while (reading < ending && *reading != '\n')
					{
					// Should limit search of ':' to one line only!!
					char *value = (char *)memchr(reading, ':', ending - reading);
					if (value == NULL)
						{
						ERRX("Missing ':' in '%.*s'", ending - reading, reading);
						return -2; // Bad Syntax!
						}
					// the tag is between reading and value.
					const struct atom_string *tag = Atomize(value - reading);
					if (tag == NULL)
						return -1; // Out of space!
					else if (tag->name[0]) // [empty name '' is "match all"]
						++reading; // Skip over ':'

					// Find the end of the header value. If the data uses
					// CR+LF to terminate lines, then the extra CR's are
					// included in the value (e.g. the value includes
					// CR+LF of continuation lines, and a trailing CR).
					char *eol = NULL;
					value = reading;
					for (;;)
						{
						eol = (char *)memchr(value, '\n', ending - value);
						if (eol == NULL)
							{
							ERRX("Cannot find newline in '%.*s'", ending - reading, reading);
							return -2; // Syntax error!
							}
						if (eol+1 == ending)
							break;
						if (eol[1] != ' ' && eol[1] != '\t')
							break;
						// Continuation line
						value = eol+1;
						}

					if (tag->fops && tag->fops->parse)
						{
						struct field_value *f = (struct field_value *)ReserveAligned(field_size);
						if (f == NULL)
							{
							ERRX("Out of memory for '%.*s'", eol - reading, reading);
							return -1; // Out of space;
							}

						// For the duration of the "parse" call, the next field
						// pointer of the new field structure points to the list of
						// previous fields for the current record. The pointer
						// is NULL, if this is the first field of the record.
						// Parse function can examine all previous fields, if
						// desires.
						f->tag = tag;
						f->next = first_field;
						f->value_int = -1;
						const int res = tag->fops->parse(*this, eol - reading, *f);
						if (res != 0)
							return res; // Out of space/syntax error
						if (f->value_int == -1)
							{
							ERRX("Parser of '%s' did't initialize value", tag->name);
							return -2;
							}
						if (last_field == NULL)
							{
							// Allocate record structure...
							struct record_value* r = (struct record_value *)ReserveAligned(sizeof(struct record_value));
							if (r == NULL)
								return -1; // Out of space!
							*last_record = r;
							r->next = NULL;
							r->fields = NULL;
							last_record = &r->next;
							last_field = &r->fields;
							first_field = f;
							}
						// Put the new field as last field
						f->next = NULL; // Remove the 'first_field' link
						*last_field = f;
						last_field = &f->next;
						}
					else
						{
						// No parsing function for this tag, ignore the field.
						reading = eol;
						}
					reading++; // Skip over value terminating '\n'
					}
				if (reading == ending || *reading == 0)
					break;
				// Skip over LF and start a new record
				++reading; // Skip over record separating '\n'
				if (body_parser)
					{
					const int res = body_parser(*this, *last_record);
					if (res < 0)
						return res;
					}
				last_field = NULL;
				first_field = NULL;
				}
			return ending - loading;
			}
		
		void *Scanner::Reserve(size_t size)
			/**
			 * Reserve space for parsed information.
			 *
			 * Reserve space for a fixed amount and return
			 * a pointer to it.
			 *
			 * @param size The size of the area to reserve.
			 * @return a pointer to the reserved header, or
			 * NULL if there is no room.
			 *
			 * WARNING: Note that there is alignment processing
			 * mixing this with the next must be done cautiosly,
			 * keeping alignment requirements in mind!
			 */
			{
			const size_t available = reading - loading;
			void *tmp = loading;
			if (available < size)
				{
				tmp = overflow.ReserveAligned(size, 0);
				if (tmp == NULL)
					return NULL;
				}
			else
				{
				loading += size;
				}
			memset(tmp, 0, size);
			return tmp;
			}

		void *Scanner::Aligned()
			/**
			 * Return aligned reference.
			 *
			 * Make sure that the next reservation starts from
			 * a memory address that can store a usable struct
			 * pointer. This function does nothing, if this
			 * aligment feature is alredy satisfied. Otherwise
			 * it reserves some space to achieve the alignment.
			 *
			 * @return Aligned pointer or NULL, if out of space.
			 */
			{
			const size_t non_aligned = (size_t)loading & align_mask;
			if (non_aligned)
				{
				void *tmp = Reserve(align_mask - non_aligned + 1);
				if (tmp == NULL)
					return NULL;
				memset(tmp, '-', align_mask - non_aligned + 1);
				}
			return loading;
			}

		void *Scanner::ReserveAligned(size_t size)
			{
			void *tmp = Aligned();
			const size_t available = reading - loading;
			if (tmp && available >= size)
				loading += size;
			else
				tmp = overflow.ReserveAligned(size, align_mask);
			if (tmp)
				memset(tmp, 0, size);
			return tmp;
			}

		void *Scanner::Reserve(size_t size, int len)
			/**
			 * Reserve space for the parsed information and move a string.
			 *
			 * Reserse space for the fixed header part and optionally
			 * move a section of the unparsed data from the 'reading'
			 * after the header. On successful operation, the 'reading'
			 * pointer advances over the moved data.
			 *
			 * Append an extra NUL byte to the copied data.
			 *
			 * @param size The size of the required header
			 * @param len Number of bytes to copy from 'reading'
			 *
			 * @return a pointer to the reserved header, or NULL if
			 * there is no room.
			 */
			{
			const size_t space = len + size + 1;
			void* tmp = NULL;
			if (size == 0 || Aligned())
				{
				const size_t available = reading - loading + len;
				if (available >= space)
					{
					tmp = loading;
					loading += space;
					}
				}
			if (tmp == NULL)
				{
				tmp = overflow.ReserveAligned(space, size ? align_mask : 0);
				if (tmp == NULL)
					return NULL;
				}
			if (size)
				memset(tmp, 0, size);
			if (len)
				memmove((char*)tmp + size, reading, len);
			((char *)tmp)[size + len] = 0;
			reading += len;
			return tmp;
			}
		
		const struct atom_string *Scanner::Atomize(int len)
			{
			// First check match against a static atom

			if (db.atoms_static)
				{
				for (int i = 0; db.atoms_static[i]; ++i)
					{
					const atom_string* s = db.atoms_static[i];
					if (s->name[0] == 0)
						{
						// Matched static atom, use it. Note: name == "" matches all, and adding such
						// entry as a last to fixed atoms with NULL fops, is an easy way to ignore all
						// unknown tag values. THE READING POINTER IS NOT ADVANCED, the fops parse will
						// receive the complete tag + value.
						return s;
						}
					else if (strncmp(s->name, reading, len) == 0 && s->name[len] == 0)
						{
						// Matched non-null static atom, use it.
						reading += len;
						return s;
						}
					}
				}

			struct atom_list** ptr = &db.atoms;
			struct atom_list* a;
			while ((a = *ptr) != NULL)
				{
				if (strncmp(a->atom->name, reading, len) == 0 && a->atom->name[len] == 0)
					{
					// An existing atom matches, no need to allocate new
					reading += len;
					return a->atom;
					}
				ptr = &a->next;
				}
			a = (struct atom_list *)ReserveAligned(sizeof(struct atom_list));
			if (a == NULL)
				return NULL;
			*ptr = a;
			a->next = NULL;
			a->atom = (struct atom_string *)Reserve(offsetof(struct atom_string, name), len);
			if (a->atom == NULL)
				return NULL;
			a->atom->fops = &fops_default;
			return a->atom;
			}

		static const char *next_wsp(const char *src, size_t len)
			{
			while (len > 0)
				if (isspace(*src))
					return src;
				else
					{
					++src;
					--len;
					}
			return NULL;
			}

		int Scanner::NextItem(const char *end, int delimiter, size_t &size, size_t& tail)
			{
			int found = 0;
			// Skip over leading white space
			while (reading < end && isspace(*reading))
				++reading;

			// If delimiter is 0, search for next white space.
			const char *mark = delimiter ?
				(const char *)memchr(reading, delimiter, end - reading) :
				next_wsp(reading, end - reading);

			if (mark)
				{
				found = 1;
				tail = 1;
				}
			else
				{
				tail = 0;
				mark = end;
				}
			// Trim trailing white space
			while (mark > reading && isspace(mark[-1]))
				{
				--mark;
				++tail;
				}
			size = mark - reading;
			return found;
			}

		int Scanner::SplitBy(int len, int delimiter, SplitCb& cb)
			{
			char *const end = reading + len;
			while (reading < end)
				{
				size_t tail;
				size_t size;
				// Delimiter not found is normal, return ignored.
				(void)NextItem(end, delimiter, size, tail);
				int res = cb.Item(*this, size);
				if (res < 0)
					return res;
				reading += tail;
				}
			return 0;
			}

		void* load_files(size_t head_room, size_t tail_room, unsigned int& flags, size_t& total, ...)
			{
			static const int MAX_FILES = 10;
			char *buf;
			
			int num_files = 0;
			const char *prefix;
			int fd[MAX_FILES];
			size_t size[MAX_FILES];
			va_list vl;

			flags = 0; // All OK for starters.

			if (SSIZE_MAX - head_room < tail_room)
				{
				ERRX("Combined head_room(%d) and tail_room(%d) exceeds SSIZE_MAX(%ld)", head_room, tail_room, SSIZE_MAX);
				return NULL;
				}
			total = head_room + tail_room;
			
			// Open the files and find out the total required buffer space.
			va_start(vl, total);
			while (num_files < MAX_FILES && (prefix = va_arg(vl, const char *)) != NULL)
				{
				const char *path = va_arg(vl, const char *);
				struct stat info;
				size[num_files] = 0;
				total += strlen(prefix);
				if (path == NULL)
					{
					fd[num_files++] = -1;
					break;
					}
				fd[num_files] = open(path, O_RDONLY);
				if (fd[num_files] == -1)
					{
					ERR("Failed opening '%s'", path);
					flags |= LOAD_FILE_NOT_FOUND;
					}
				else if (fstat(fd[num_files], &info))
					{
					close(fd[num_files]);
					fd[num_files] = -1;
					ERR("Failed stat on %s", path);
					flags |= LOAD_READ_ERROR;
					}
				else if (info.st_size > (off_t)(SSIZE_MAX - total))
					{
					close(fd[num_files]);
					fd[num_files] = -1;
					ERRX("File size would exceed maximum for '%s'", path);
					flags |= LOAD_READ_ERROR;
					}
				else
					{
					size[num_files] = (size_t)info.st_size;
					DBG(warnx("File size %u for '%s'", size[num_files], path));
					}
				total += size[num_files];
				++num_files;
				}
			va_end(vl);
			
			buf = (char *)malloc(total);
			if (buf == NULL)
				{
				ERR("Allocating buffer size=%u failed", total);
				}
			else
				{
				DBG(warnx("Allocated buffer size=%u", total));
				
				size_t fill = head_room;
				memset(buf, 0, head_room);
				
				// Load files...
				va_start(vl, total);
				for (int i = 0; i < num_files; ++i)
					{
					prefix = va_arg(vl, const char *);
					const char *path = va_arg(vl, const char *);
					const int len = strlen(prefix);
					if (len > 0)
						{
						memcpy(buf+fill, prefix, len);
						fill += len;
						}
					if (fd[i] != -1)
						{
						const ssize_t actual = read(fd[i], buf+fill, size[i]);
						if (actual < 0)
							{
							ERR("Failed to read '%s'", path);
							flags |= LOAD_READ_ERROR;
							}
						else if ((size_t)actual != size[i])
							{
							ERRX("Bad read on '%s' -- only got %d out of %d", path, actual, size[i]);
							flags |= LOAD_READ_ERROR;
							}
						else
							{
							//if (!verify_signature(path, buf+fill, size[i]))
							//{
							//	ERRX("File '%s' fails signature", path);
							//	flags |= LOAD_SIGNATURE_FAIL;
							//
							//}
							flags |= LOAD_NO_SIGNATURE; // For now, no files verified!
							fill += actual;
							}
						}
					}
				va_end(vl);
				}
			
			while (--num_files >= 0)
				if (fd[num_files] >= 0)
					close(fd[num_files]);
			return buf;
			}
		}
	}
