/*
 * This file is part of AEGIS
 *
 * Copyright (C) 2009 Nokia Corporation
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2.1 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 *
 * Author: Markku Savela
 */

/*
 * Low level parser for general "internet header" format.
 */
#ifndef _RESTOK_DB_
#define _RESTOK_DB_

#include <stddef.h> // ..for NULL

namespace RESTOK
	{
	namespace DB
		{

		class Scanner;

		class Overflow
			{
		public:
			virtual ~Overflow() {}
			virtual void* ReserveAligned(size_t, unsigned int)  { return NULL; }
			};

		/*
		 * The bit flags for the load information
		 *
		 * The bits are set when the corresponding situation
		 * has occurred for one or more files in (load_files).
		 */
		static const unsigned int
			LOAD_FILE_NOT_FOUND = 1,// File path does not exist
			LOAD_READ_ERROR = 2,	// File size or content not available
			LOAD_NO_SIGNATURE = 4,	// File is not signed
			LOAD_SIGNATURE_FAIL = 8;// File was signed, but signature failed.

		/*!
		 * The function for actual printing.
		 *
		 * @param format Standard 'printf' format string
		 * @return Number of characters printed.
		 */
		typedef int (* PrintF)(const char *format, ...);
		/*!
		 * The parse function for the tag value.
		 *
		 * Note: the value (len) includes the tag, if the match was based
		 * on a predefined tag value with NULL string (= match any).
		 *
		 * @param scanner The parsing state
		 * @param len The length of the value
		 * @param field The field structure for the current value.
		 *
		 * @return 0, for success; < 0 for any error.
		 */
		typedef int (* Parse)(Scanner &scanner, int len, struct field_value& field);
		/*!
		 * The parse function for the body following the header.
		 *
		 * Called after the header fields terminating sequence ([CR]LF[CR]LF) has been
		 * passed over. The Body function advances the scanner over the body (if any).
		 *
		 * @param scanner The parsing state
		 * @param record The header fields for the body (NULL, if NONE!)
		 * @return 0, for success; < 0 for any error.
		 */
		typedef int (* Body)(Scanner &scanner, const struct record_value* record);

		/*!
		 * Print field value.
		 *
		 * @param field The field to print.
		 * @param printer The printing function.
		 * @return Number of characters printed.
		 */
		typedef int (* Print)(const struct field_value& field, PrintF printer);

		struct field_ops
			{
			Parse parse; ///< Parsing function for the field content
			Print print; ///< Print function for the parsed field
			};

		struct atom_string
			{
			struct field_ops *fops;
			char name[];
			};

		// This stupid hack is required, because the g++ doesn't allow
		// static initialization of char[], that is you cannot do
		//	struct atom_string foo = {NULL, "foo"; }
		// which is perfecly valid in C/gcc.
#		define DB_ATOM_STATIC(_label, _fops, _name) \
			static struct {DB::field_ops* fops; char name[sizeof(_name)+1]; } _label = {_fops, _name}

		struct atom_list
			{
			struct atom_list* next;
			struct atom_string* atom;
			};

		struct field_value
			{
			const atom_string* tag;
			struct field_value* next;
			union
				{
				void* value_ptr;
				int value_int;
				};
			};
		
		struct record_value
			{
			struct record_value* next;
			struct field_value* fields;
			};
		
		struct db_header
			{
			/*!
			 * Parse data base in place.
			 *
			 * @param size The length of the string to parse.
			 * @param offset Start of the string in base (e.g. first byte to parse is base[offset])
			 * @param atoms NULL terminated array of predefined tags (pointers to atom_string)
			 * @param user_data The user data for the parser callbacks (inside Scanner class)
			 */
			int Load(int size, int offset, const struct atom_string* atoms[], void *user_data, Body body_parser, Overflow& overflow);
			/*!
			 * Dump database.
			 *
			 * If the parse/print routines desire, the resulting dump
			 * can replicate exact original data that was loaded.
			 */
			int Dump(PrintF printer) const;

			/*!
			 * The current list of dynamically defined tags.
			 *
			 * The list includes an instance for each unique tag detected.
			 */
			struct atom_list* atoms;
			/*!
			 * The predefined list of tags provided by the user.
			 */
			const struct atom_string** atoms_static;
			/*!
			 * The list of parsed records.
			 */
			struct record_value *records;
			/*
			 * The parsed binary database allocation begins.
			 */
			char base[];
			};

		class SplitCb
			{
		public:
			/*!
			 * Called for each extracted item in a Split operation.
			 *
			 * @param scanner The parsing context.
			 * @param len The length of the item (starting from scanner.reading.
			 * @return 0 if ok, < 0 for any error.
			 *
			 * After return the parsing continues at whatever the scanner.reading
			 * points -- the function should normally advance the reading point
			 * by len (e.g. scanner.reading += len).
			 */
			virtual int Item(Scanner& scanner, size_t len) = 0;
			};

		class Scanner
			{
			void* Aligned();
		public:
			Scanner(struct db_header& db, int size, int offset, void* user_data, Overflow& overflow) :
				db(db),
				overflow(overflow),
				loading(db.base),
				reading(db.base + offset),
				ending(reading + size),
				user_data(user_data)
				{}
			
			struct db_header &db;
			Overflow& overflow;

			char* loading; // The current writing point
			char* reading; // The current reading point
			char* ending; // End of input data.

			void* user_data;
			
			int Load(const struct atom_string *atoms[], Body body_parser);
			void* Reserve(size_t header, int len);
			void* ReserveAligned(size_t header);
			void* Reserve(size_t bytes);
			const struct atom_string* Atomize(int len);

			/*!
			 * Extract next item from input.
			 *
			 * @param end End of input range
			 * @param delimiter The item delimiter (or 0)
			 * @retval size The the length of the item (can be zero!)
			 * @retval tail The item separator and possible trailing white space length
			 * @return 1, if delimiter found, and 0 otherwise
			 *
			 * The search for the item starts from the current 'reading' pointer and
			 * terminates either on delimiter or at the end of the range.
			 *
			 * First the search advances the reading point over any white space (WSP)
			 * WITHOUT testing the delimiter (significant if delimiter is white space).
			 *
			 * If delimiter == 0, then any white space (isspace) acts as a delimiter.
			 *
			 * Then delimiter is located and any preceding and trailing white space
			 * is trimmed away and included into tail count.
			 *
			 * If the delimiter is not found, the search terminates at end, and the
			 * trailing white space preceding the end is returned as tail.
			 *
			 * @code
			 * // before:
			 * // |scanner.reading
			 * // |                                   |end
			 * // |                                   |
			 * // WSP itemstring WSP delimiter WSP ...|
			 * // after:                          |   |
			 * //     |--size---|-----tail--------|   |
			 * //     |scanner.reading            |   |end
			 * @endcode
			 * 
			 * 
			 */
			int NextItem(const char *end, int delimiter, size_t& size, size_t& tail);

			/*!
			 * Split input section by a delimiter.
			 *
			 * Convenience function which uses the NextItem repeatedly,
			 * and calls the provided callback for each extracted
			 * item.
			 *
			 * @param len The length of the input region (starting from Scanner::reading)
			 * @param delimiter The item delimiter (or 0)
			 * @param cb The callback class.
			 * @return 0 on ok, < 0 for error.
			 */
			int SplitBy(int len, int delimiter, SplitCb& cb);
			};

		/*!
		 * Load a set of files into a allocated buffer.
		 *
		 * @param head_room The headroom in the buffer.
		 * @param tail_room The extra space at end in buffer.
		 * @retval info Returns non-zero bitmask (load information) on trouble detected
		 * @retval total The total size of the loaded files (including head_room, tail_room and prefixes).
		 *
		 * A list of <prefix, path> entries follows the fixed parameters, both
		 * are "const char*" values.
		 *
		 * A NULL 'prefix' or 'path' terminates the list. A terminating non-NULL
		 * 'prefix' with NULL path can be used to add a trailing constant string
		 * the final data.
		 *
		 * The final buffer contains the concatenation of the prefix strings
		 * and file contents.
		 */
		extern void* load_files(size_t head_room, size_t tail_room, unsigned& info, size_t& total, ...);
		}
	}

#endif
