/*
 * This file is part of AEGIS
 *
 * Copyright (C) 2010 Nokia Corporation
 *
 * Contact: Laszlo Papp <ext-laszlo.papp@nokia.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 *
 * Author: Laszlo Papp
 */

#include <sys/socket.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <linux/capability.h>
#include <netinet/in.h>
#include <errno.h>
#include <unistd.h>
#include <stdbool.h>

#include <linux/sockios.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include "socketpolicy.h"

#ifdef HAVE_LINUX_AEGIS_SOCKETSCTL_H
#include <linux/aegis/socketsctl.h>

#define SECUREFS_PATH "/sys/kernel/security"

static const char *const lsm_socket_file = SECUREFS_PATH"/"LSM_SOCKETS_SECURITY_DIR"/"LSM_SOCKETS_SECURITY_FILE;

#else

struct skt_params {
	__u32 family;
	__u32 type;
	__u32 protocol;
	__s32 creds_type;
	__s32 creds_value;
};
#endif

struct skt_ioc_args {
    uint32_t size;
    struct skt_params *params;
};

#define DEFAULT_SOCKET_POLICY_NUM 20
#define SOCKET_POLICY_PARAMS_SIZE 20
#define ARRAY_SIZE(a) (sizeof (a) / sizeof ((a)[0]))


/**
 * struct type_str: Holds constant string and length
 * @str: The string
 * @type: The type
 */
struct type_str {
   const char *const str;
   int type;
};
#define TYPE_STR(s, t) {(s), (t)}

/*
 * Static mapping of type names <-> values
 *
 * Other userspace applcation may require similar need. The array of
 * type names should be provided globally by a standard header file
 */
static const struct type_str type_names[] = {
   TYPE_STR("stream", SOCK_STREAM),
   TYPE_STR("dgram", SOCK_DGRAM),
   TYPE_STR("raw", SOCK_RAW),
   TYPE_STR("rdm", SOCK_RDM),
   TYPE_STR("seqpacket", SOCK_SEQPACKET),
   TYPE_STR("packet", SOCK_PACKET),
};
#define TYPE_NAMES_SIZE ARRAY_SIZE(type_names)

/**
 * struct protocol_str: Holds constant string and length
 * @str: The string
 * @protocol: The protocol
 */
struct protocol_str {
   const char *const str;
   int protocol;
};
#define PROTOCOL_STR(s, p) {s, p}

/*
 * Static mapping of protocol names <-> values
 *
 * Other userspace applcation may require similar need. The array of
 * protocol names should be provided globally by a standard header file
 */
static const struct protocol_str ip_protocol_names[] = {
   PROTOCOL_STR("stream", IPPROTO_IP),
   PROTOCOL_STR("hopopts", IPPROTO_HOPOPTS),
   PROTOCOL_STR("icmp", IPPROTO_ICMP),
   PROTOCOL_STR("igmp", IPPROTO_IGMP),
   PROTOCOL_STR("ipip", IPPROTO_IPIP),
   PROTOCOL_STR("tcp", IPPROTO_TCP),
   PROTOCOL_STR("egp", IPPROTO_EGP),
   PROTOCOL_STR("pup", IPPROTO_PUP),
   PROTOCOL_STR("udp", IPPROTO_UDP),
   PROTOCOL_STR("idp", IPPROTO_IDP),
   PROTOCOL_STR("tp", IPPROTO_TP),
   PROTOCOL_STR("ipv6", IPPROTO_IPV6),
   PROTOCOL_STR("routing", IPPROTO_ROUTING),
   PROTOCOL_STR("fragment", IPPROTO_FRAGMENT),
   PROTOCOL_STR("rsvp", IPPROTO_RSVP),
   PROTOCOL_STR("gre", IPPROTO_GRE),
   PROTOCOL_STR("esp", IPPROTO_ESP),
   PROTOCOL_STR("ah", IPPROTO_AH),
   PROTOCOL_STR("icmpv6", IPPROTO_ICMPV6),
   PROTOCOL_STR("none", IPPROTO_NONE),
   PROTOCOL_STR("dstopts", IPPROTO_DSTOPTS),
   PROTOCOL_STR("mtp", IPPROTO_MTP),
   PROTOCOL_STR("encap", IPPROTO_ENCAP),
   PROTOCOL_STR("pim", IPPROTO_PIM),
   PROTOCOL_STR("comp", IPPROTO_COMP),
   PROTOCOL_STR("sctp", IPPROTO_SCTP),
   PROTOCOL_STR("raw", IPPROTO_RAW),
   PROTOCOL_STR("max", IPPROTO_MAX),
   /* End of table indicated by a NULL str */
   PROTOCOL_STR(NULL, 0),
};

/**
 * struct family_str: Holds constant string and length
 * @str: The string
 * @len: The length
 */
struct family_str {
   const char *const str;
   int family;
   const struct protocol_str *protocols;
};
#define FAMILY_STR(s, f, p) {s, f, p}

/*
 * Static mapping of family names <-> values
 *
 * Other userspace applcation may require similar need. The array of
 * protocol names should be provided globally by a standard header file
 */
static const struct family_str family_names[] = {
    FAMILY_STR("unspec", AF_UNSPEC, NULL),
    FAMILY_STR("local", AF_LOCAL, NULL),
    FAMILY_STR("unix", AF_LOCAL, NULL),
    FAMILY_STR("file", AF_LOCAL, NULL),
    FAMILY_STR("inet", AF_INET, ip_protocol_names),
    FAMILY_STR("ax25", AF_AX25, NULL),
    FAMILY_STR("ipx", AF_IPX, NULL),
    FAMILY_STR("appletalk", AF_APPLETALK, NULL),
    FAMILY_STR("netrom", AF_NETROM, NULL),
    FAMILY_STR("bridge", AF_BRIDGE, NULL),
    FAMILY_STR("atmpvc", AF_ATMPVC, NULL),
    FAMILY_STR("x25", AF_X25, NULL),
    FAMILY_STR("inet6", AF_INET6, ip_protocol_names),
    FAMILY_STR("rose", AF_ROSE, NULL),
    FAMILY_STR("decnet", AF_DECnet, NULL),
    FAMILY_STR("netbeui", AF_NETBEUI, NULL),
    FAMILY_STR("security", AF_SECURITY, NULL),
    FAMILY_STR("key", AF_KEY, NULL),
    FAMILY_STR("netlink", AF_NETLINK, NULL),
    FAMILY_STR("route", AF_ROUTE, NULL),
    FAMILY_STR("packet", AF_PACKET, NULL),
    FAMILY_STR("ash", AF_ASH, NULL),
    FAMILY_STR("econet", AF_ECONET, NULL),
    FAMILY_STR("sna", AF_SNA, NULL),
    FAMILY_STR("irda", AF_IRDA, NULL),
    FAMILY_STR("pppox", AF_PPPOX, NULL),
    FAMILY_STR("wanpipe", AF_WANPIPE, NULL),
    FAMILY_STR("atmsvc", AF_ATMSVC, NULL),
    FAMILY_STR("bluetooth", AF_BLUETOOTH, NULL),
    FAMILY_STR("iucv", AF_IUCV, NULL),
    FAMILY_STR("rxrpc", AF_RXRPC, NULL),
    FAMILY_STR("isdn", AF_ISDN, NULL),
    FAMILY_STR("phonet", AF_PHONET, NULL),
    FAMILY_STR("ieee802154", AF_IEEE802154, NULL),
    FAMILY_STR("max", AF_MAX, NULL),
};
#define FAMILY_NAMES_SIZE ARRAY_SIZE(family_names)

/**
 * socket_family2str - Convert the family number to a family name
 *
 * @family: The family number
 *
 * Return the family name (including the terminating
 * NULL byte). If the return value is larger than str_len, then
 * the name could not be returned fully.
 */
const char *const socket_family2str(long family)
{
    int i;
    for (i = 0; i < FAMILY_NAMES_SIZE; ++i) {
        if (family == family_names[i].family)
            return family_names[i].str;
    }
    return 0;
}

/**
 * socket_str2family - Convert the family name to a family number
 *
 * @str: The name of the family
 * @str_len: The length of the family name.
 *
 * Return the index to family_names or -EINVAL, if conversion fails.
 */
static int socket_str2family(const char *str, size_t str_len)
{
    int family;
    for (family = 0; family < FAMILY_NAMES_SIZE; ++family) {
        if (str_len == strlen(family_names[family].str) &&
                memcmp(family_names[family].str, str, str_len) == 0) {
            return family;
        }
    }
    return -EINVAL;
}

/**
 * socket_type2str - Convert type number to type name
 *
 * @type: The type number
 *
 * Return the type name (including the terminating
 * NULL byte). If the return value is larger than str_len, then
 * the name could not be returned fully.
 */
const char *const socket_type2str(long type)
{
    int i;
    for (i = 0; i < TYPE_NAMES_SIZE; ++i) {
        if (type == type_names[i].type)
            return type_names[i].str;
    }
    return 0;
}

/**
 * socket_str2type - Convert the type name to a type number
 *
 * @str: The name of the type
 * @str_len: The length of the type name.
 *
 * Return the type number or -EINVAL, if conversion fails.
 */
static long socket_str2type(const char *str, size_t str_len)
{
    int i;
    for (i = 0; i < TYPE_NAMES_SIZE; ++i) {
        if (str_len == strlen(type_names[i].str) &&
                memcmp(type_names[i].str, str, str_len) == 0) {
            return type_names[i].type;
        }
    }
    return -EINVAL;
}

/**
 * socket_protocol2str - Convert protocol number to protocol name
 *
 * @protocol: The protocol number
 * @names: The procotol names table
 *
 * Return the protocol name (including the terminating
 * NULL byte). If the return value is larger than str_len, then
 * the name could not be returned fully.
 */
const char *const socket_protocol2str(long protocol, const struct protocol_str *names)
{
    int i;
    if (names) {
        for (i = 0; names[i].str; ++i) {
            if (protocol == names[i].protocol)
                 return names[i].str;
	}
    }
    return 0;
}

/**
 * socket_str2protocol - Convert the protocol name to a protocol number
 *
 * @str: The name of the protocol
 * @str_len: The length of the protocol name.
 * @names: The procotol names table
 *
 * Return the protocol number or -EINVAL, if conversion fails.
 */
static long socket_str2protocol(const char *str, size_t str_len, const struct protocol_str *names)
{
    int protocol;
    if (!names)
        return -EINVAL;

    for (protocol = 0; names[protocol].str; ++protocol) {
        if (str_len == strlen(names[protocol].str) &&
                memcmp(names[protocol].str, str, str_len) == 0) {
            return names[protocol].protocol;
        }
    }
    return -EINVAL;
}

/**
 * socket_policy_start()
 *
 * This function is to be called to allocate a default memory area for the socket policies
 *
 * Notes
 * Return a pointer to the allocated memory. On error, these functions return NULL.
 */
struct skt_ioc_args *socket_policy_start()
{
    /* printf("SKT_PARAMS: %li\r\n", sizeof(struct skt_params)); */
    struct skt_ioc_args *args = (struct skt_ioc_args *)calloc(1, sizeof(struct skt_ioc_args));
    args->params = (struct skt_params *)calloc(DEFAULT_SOCKET_POLICY_NUM, sizeof(struct skt_params));
    if (args->params)
        return args;
    free(args);
    return NULL;
}

/**
 * socket_policy_add() - Add a new socket policy that will be loaded into the kernel
 * @handle:   contains the handle structure for the socket and it will be modified
 * @socketstr: contains the socket family, type and protocol values in one string with separator (' ')
 * @cred_type: contains the credential the required type
 * @cred_value: contains the credential the required value
 *
 * Return 0 if the socket and credential parameters could be added successfully
 */
int socket_policy_add(struct skt_ioc_args *handle, const char *socketstr, int cred_type, int cred_value)
{
    const char *p = 0;
    size_t l = 0;
    int family_index;
    int type = 0;
    int protocol = 0;

    if (handle->size >= DEFAULT_SOCKET_POLICY_NUM) {
        struct skt_params *tmp = realloc(handle->params, (handle->size + 1) * sizeof(struct skt_params));
	if (!tmp)
            return -ENOMEM;
        handle->params = tmp;
    }

    p = socketstr;
    l = strcspn(p, " ");
    family_index = socket_str2family(p, l);
    if (family_index < 0)
        return -EINVAL;
    handle->params[handle->size].family = family_names[family_index].family;

    p += l;
    if (*p) {
        p += 1;
        l = strcspn(p, " ");
        type = socket_str2type(p, l);
        if (type < 0)
            return -EINVAL;
        p += l;
    }
    handle->params[handle->size].type = type;

    if (*p) {
        p += 1;
        l = strcspn(p, " ");
        protocol = socket_str2protocol(p, l, family_names[family_index].protocols);
        if (protocol < 0)
            return -EINVAL;
    }
    handle->params[handle->size].protocol = protocol;
    handle->params[handle->size].creds_type = cred_type;
    handle->params[handle->size].creds_value = cred_value;

    ++handle->size;
    return 0;
}

#ifdef HAVE_LINUX_AEGIS_SOCKETSCTL_H
/**
 * socket_policy_finish() - Finish the socket policy addition session
 * @handle:    contains the handle structure for the socket that is loaded into the kernel
 *
 * Return 0 if the socket policy could be loaded into the kernel module
 */
int socket_policy_finish(struct skt_ioc_args *handle)
{
    int fd = -1;
    uint32_t shandle = 0;
    int i;
    char *ioctl_arg = 0;
    int ret = 0;

    if (!handle)
        return -EINVAL;
	
    fd = open(lsm_socket_file, 0);
    if (fd == -1) {
       ret = -errno;
       goto out;
    }

    shandle = handle->size * sizeof(struct skt_params) + sizeof(handle->size);

    ioctl_arg = (char *)malloc(sizeof(struct skt_params) * handle->size + sizeof(uint32_t));
    if (!ioctl_arg) {
        ret = -ENOMEM;
	goto out;
    }
    memcpy(ioctl_arg, &shandle, sizeof(uint32_t));
    for (i = 0; i < handle->size; ++i)
        memcpy(ioctl_arg + sizeof(uint32_t) + i*SOCKET_POLICY_PARAMS_SIZE, &handle->params[i], SOCKET_POLICY_PARAMS_SIZE);

    if (ioctl(fd, SIOCLSMSKT_LOAD, ioctl_arg) == -1)
        ret = -EINVAL;

out:
    if (fd >= 0)
        close(fd);
    if (ioctl_arg)
        free(ioctl_arg);
    socket_policy_abort(handle);
    return ret;
}
#else
/**
 * socket_policy_finish() - Dummy Finish
 * @handle:    contains the handle structure for the socket that is loaded into the kernel
 *
 * Return 0
 */
int socket_policy_finish(struct skt_ioc_args *handle)
{
    socket_policy_abort(handle);
    return 0;
}
#endif

/**
 * socket_policy_abort() - Abort the socket policy addition or starting session
 * @handle:    contains the handle structure for the socket that will be deallocated
 */
void socket_policy_abort(struct skt_ioc_args *handle)
{
    if (!handle)
        return;
    if (handle->params)
        free(handle->params);
    free(handle);
}

