/*
 * This file is part of AEGIS
 *
 * Copyright (C) 2009 Nokia Corporation
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * version 2.1 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 *
 * Author: Markku Savela
 */

#include <stddef.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#include "restok_def.h"
#include "restok_db.h"
#include "restok_cnf.h"
#include "status_cnf.h"

namespace STATUS
	{
	using namespace RESTOK;

	class Allocator : public DB::Overflow
		{
	public:
		Allocator(RESTOK::configuration& cnf) : cnf(cnf) {}
		void *ReserveAligned(size_t size, unsigned int mask);
	private:
		RESTOK::configuration& cnf;
		};

	void *Allocator::ReserveAligned(size_t size, unsigned int mask)
		{
		return cnf.alloc(size, mask);
		}


	static int ParsePackage(struct DB::Scanner &scanner, int len, struct DB::field_value& field)
		{
		struct configuration & status = *(struct configuration*)scanner.user_data;
		char *const end = scanner.reading + len;
		if (field.next != NULL)
			{
			warnx("The tag '%.*s' was not the first in record", len, scanner.reading);
			return -2;
			}

		size_t tail;
		size_t size;
		(void)scanner.NextItem(end, '\n', size, tail);
		if (size <= 0 || scanner.reading + size + tail != end)
			{
			warnx("The tag '%.*s' has bad syntax", len, scanner.reading);
			return -2;
			}

		status.package = status.cnf->find_package(scanner.reading, size);
		if (status.package == NULL)
			{
			status.package = status.cnf->make_package(scanner, size);
			if (status.package == NULL)
				{
				warnx("Out of 'status' memory while on Pacakge '%.*s'", size, scanner.reading);
				return -1; // Out of memory
				}
			}

		// For all packages mentioned in dpkg status, set the source/origin
		// if not defined already...
		if (status.package->source == NULL)
			status.package->source = status.source;

		// Remember the package for status processing
		field.value_ptr = status.package;
		scanner.reading = end;
		return 0;
		}

	static int PrintPackage(const struct DB::field_value& field, DB::PrintF printer)
		{
		struct package *pn = (struct package *)field.value_ptr;
		return pn ? printer("%s", pn->name) : 0;
		}


	class StatusCb : public DB::SplitCb
		{
	public:
		struct RESTOK::package& pkg;
		unsigned int status_bits;
		StatusCb(struct RESTOK::package& pkg) : SplitCb(), pkg(pkg), status_bits(0) {}
		int Item(DB::Scanner& scanner, size_t size);

		};

	enum
		{
		PkgStatus_ok		= 1,
		PkgStatus_install	= 1 << 1,
		PkgStatus_installed	= 1 << 2,
		PkgStatus_not_installed	= 1 << 3,
		PkgStatus_purge		= 1 << 4,
		PkgStatus_deinstall	= 1 << 5,
		PkgStatus_config_files	= 1 << 6,
		};

	static const unsigned int PKG_INSTALLED = (PkgStatus_installed);
	static const unsigned int PKG_NOT_INSTALLED = (PkgStatus_not_installed);

#	define STATUS(s,n) {s, sizeof(s)-1, PkgStatus_##n}

	static const struct { const char* str; size_t len; unsigned int bit; }
	flags[] =
		{
		STATUS("ok", ok),
		STATUS("install", install),
		STATUS("installed", installed),
		STATUS("not-installed", not_installed),
		STATUS("purge", purge),
		STATUS("deinstall", deinstall),
		STATUS("config-files", config_files),
		};

	int StatusCb::Item(DB::Scanner& scanner, size_t size)
		{
		for (size_t i = 0; i < sizeof(flags) / sizeof(flags[0]); ++i)
			{
			if (size == flags[i].len && memcmp(scanner.reading, flags[i].str, size) == 0)
				{
				scanner.reading += size;
				status_bits |= flags[i].bit;
				return 0;
				}
			}
		warnx("Unkown status flag '%.*s'", size, scanner.reading);
		// Just a warning -- ignore unknown status symbols
		return 0;
		}

	static int ParseStatus(struct DB::Scanner &scanner, int len, struct DB::field_value& field)
		{
		struct configuration & status = *(struct configuration*)scanner.user_data;
		field.value_ptr = status.package;
		if (status.package)
			{
			StatusCb cb(*status.package);
			const int ret = scanner.SplitBy(len, ' ', cb);
			if (ret < 0)
				return ret;

			if (cb.status_bits & PKG_INSTALLED)
				{
				// Package is a real installed package (not
				// alias or other Aegis generated pseudo-package).
				// Mark it as such by setting the 'alias' to
				// point self.
				status.package->alias = status.package;
				}
			else if (cb.status_bits & PKG_NOT_INSTALLED)
				{
				// Package is not installed. Mark it "undefined"
				// by clearing the 'alias' field.
				status.package->alias = NULL;
				}
			return 0;
			}
		warnx("Status tag without package");
		return -2;
		}

	static int PrintStatus(const struct DB::field_value& field, DB::PrintF printer)
		{
		return 0;
		}


	static struct DB::field_ops fops_status =
		{
		ParseStatus,
		PrintStatus,
		};

	static struct DB::field_ops fops_package =
		{
		ParsePackage,
		PrintPackage,
		};

	DB_ATOM_STATIC(ignore, NULL, "");
	DB_ATOM_STATIC(status, &fops_status, "Status");
	DB_ATOM_STATIC(package, &fops_package, "Package");

	static const DB::atom_string* atom_list[] =
		{
		(DB::atom_string *)&package,
		(DB::atom_string *)&status,
		(DB::atom_string *)&ignore, // Include this to ignore all other fields.
		NULL,
		};

	struct configuration* configuration::merge(const char* origin, RESTOK::configuration* cnf)
		{
		size_t size;
		unsigned int flags;

		configuration *status = (configuration *)DB::load_files
			(offsetof(struct configuration, data), 0, flags, size,
			"", "/var/lib/dpkg/status",
			NULL);

		if (status == NULL)
			{
			ERRX("Allocating buffer size=%d failed", size);
			return NULL;
			}

		status->cnf = cnf;

		Allocator overflow(*cnf);

		if (origin && cnf->pkg_domains)
			{
			const size_t size = strlen(origin);
			status->source = cnf->find_resource(cnf->pkg_domains, origin, strlen(origin));
			if (status->source == NULL)
				{
				// Unknown source, create entry... note +1 is for the NUL byte.
				status->source = (struct RESTOK::atom *)cnf->alloc
					(offsetof(struct RESTOK::atom, token) + size + 1, 7);
				if (status->source == NULL)
					{
					ERRX("Cannot allocate origin '%s'", origin);
					status->release();
					return NULL;
					}
				memcpy(status->source->token, origin, size+1);
				status->source = RESTOK::configuration::init_resource(status->source, size, *cnf->pkg_domains);
				warnx("Source name %s", status->source->token);
				}
			}

		int len = status->db.Load(
			size - offsetof(struct configuration, data),
			status->data - status->db.base,
			atom_list, status,
			NULL,
			overflow);
		if (len < 0)
			{
			ERRX("Error merging the status");
			status->release();
			return NULL;
			}
		DBG(warnx("Load returned %d\n", len));
		return status;
		}

	void configuration::release()
		{
		restok_cnf_free(cnf);
		free((void*)this);
		}
	}
